"""
The help command. The basic idea is that help texts for commands are best
written by those that write the commands - the developers. So command-help is
all auto-loaded and searched from the current command set. The normal,
database-tied help system is used for collaborative creation of other help
topics such as RP help or game-world aides. Help entries can also be created
outside the game in modules given by ``settings.FILE_HELP_ENTRY_MODULES``.

"""

from collections import defaultdict
from dataclasses import dataclass
from itertools import chain

from django.conf import settings

from evennia.help.filehelp import FILE_HELP_ENTRIES
from evennia.help.models import HelpEntry
from evennia.help.utils import help_search_with_index, parse_entry_for_subcategories
from evennia.utils import create, evmore
from evennia.utils.ansi import ANSIString
from evennia.utils.eveditor import EvEditor
from evennia.utils.utils import (
    class_from_module,
    dedent,
    format_grid,
    inherits_from,
    pad,
)

CMD_IGNORE_PREFIXES = settings.CMD_IGNORE_PREFIXES
COMMAND_DEFAULT_CLASS = class_from_module(settings.COMMAND_DEFAULT_CLASS)
HELP_MORE_ENABLED = settings.HELP_MORE_ENABLED
DEFAULT_HELP_CATEGORY = settings.DEFAULT_HELP_CATEGORY
HELP_CLICKABLE_TOPICS = settings.HELP_CLICKABLE_TOPICS

# limit symbol import for API
__all__ = ("CmdHelp", "CmdSetHelp")


@dataclass
class HelpCategory:
    """
    Mock 'help entry' to search categories with the same code.

    """

    key: str

    @property
    def search_index_entry(self):
        return {
            "key": self.key,
            "aliases": "",
            "category": self.key,
            "no_prefix": "",
            "tags": "",
            "text": "",
        }

    def __hash__(self):
        return hash(id(self))


class CmdHelp(COMMAND_DEFAULT_CLASS):
    """
    Get help.

    Usage:
      help
      help <topic, command or category>
      help <topic>/<subtopic>
      help <topic>/<subtopic>/<subsubtopic> ...

    Use the 'help' command alone to see an index of all help topics, organized
    by category. Some big topics may offer additional sub-topics.

    """

    key = "help"
    aliases = ["?"]
    locks = "cmd:all()"
    arg_regex = r"\s|$"

    # this is a special cmdhandler flag that makes the cmdhandler also pack
    # the current cmdset with the call to self.func().
    return_cmdset = True

    # Help messages are wrapped in an EvMore call (unless using the webclient
    # with separate help popups) If you want to avoid this, simply add
    # 'HELP_MORE_ENABLED = False' in your settings/conf/settings.py
    help_more = HELP_MORE_ENABLED

    # colors for the help index
    index_type_separator_clr = "|w"
    index_category_clr = "|W"
    index_topic_clr = "|G"

    # suggestion cutoff, between 0 and 1 (1 => perfect match)
    suggestion_cutoff = 0.6

    # number of suggestions (set to 0 to remove suggestions from help)
    suggestion_maxnum = 5

    # separator between subtopics:
    subtopic_separator_char = r"/"

    # should topics disply their help entry when clicked
    clickable_topics = HELP_CLICKABLE_TOPICS

    def msg_help(self, text, **kwargs):
        """
        messages text to the caller, adding an extra oob argument to indicate
        that this is a help command result and could be rendered in a separate
        help window.

        """
        if type(self).help_more:
            usemore = True

            if self.session and self.session.protocol_key in (
                "webclient/websocket",
                "webclient/ajax",
            ):
                try:
                    options = self.account.db._saved_webclient_options
                    if options and options["helppopup"]:
                        usemore = False
                except KeyError:
                    pass

            if usemore:
                # adding the 'text_kwargs' keyword means it will be sent with the text outputfunc
                # for every page.
                evmore.msg(
                    self.caller, text, session=self.session, text_kwargs={"type": "help"}, **kwargs
                )
                return

        self.msg(text=(text, {"type": "help"}))

    def format_help_entry(
        self,
        topic="",
        help_text="",
        aliases=None,
        suggested=None,
        subtopics=None,
        click_topics=True,
    ):
        """This visually formats the help entry.
        This method can be overridden to customize the way a help
        entry is displayed.

        Args:
            title (str, optional): The title of the help entry.
            help_text (str, optional): Text of the help entry.
            aliases (list, optional): List of help-aliases (displayed in header).
            suggested (list, optional): Strings suggested reading (based on title).
            subtopics (list, optional): A list of strings - the subcategories available
                for this entry.
            click_topics (bool, optional): Should help topics be clickable. Default is True.

        Returns:
            help_message (str): Help entry formated for console.

        """
        separator = "|C" + "-" * self.client_width() + "|n"
        start = f"{separator}\n"

        title = f"|CHelp for |w{topic}|n" if topic else "|rNo help found|n"

        if aliases:
            aliases = " |C(aliases: {}|C)|n".format("|C,|n ".join(f"|w{ali}|n" for ali in aliases))
        else:
            aliases = ""

        help_text = "\n" + dedent(help_text.strip("\n")) if help_text else ""

        if subtopics:
            if click_topics:
                subtopics = [
                    f"|lchelp {topic}/{subtop}|lt|w{topic}/{subtop}|n|le" for subtop in subtopics
                ]
            else:
                subtopics = [f"|w{topic}/{subtop}|n" for subtop in subtopics]
            subtopics = "\n|CSubtopics:|n\n  {}".format(
                "\n  ".join(
                    format_grid(
                        subtopics, width=self.client_width(), line_prefix=self.index_topic_clr
                    )
                )
            )
        else:
            subtopics = ""

        if suggested:
            suggested = sorted(suggested)
            if click_topics:
                suggested = [f"|lchelp {sug}|lt|w{sug}|n|le" for sug in suggested]
            else:
                suggested = [f"|w{sug}|n" for sug in suggested]
            suggested = "\n|COther topic suggestions:|n\n{}".format(
                "\n  ".join(
                    format_grid(
                        suggested, width=self.client_width(), line_prefix=self.index_topic_clr
                    )
                )
            )
        else:
            suggested = ""

        end = start

        partorder = (start, title + aliases, help_text, subtopics, suggested, end)

        return "\n".join(part.rstrip() for part in partorder if part)

    def format_help_index(
        self, cmd_help_dict=None, db_help_dict=None, title_lone_category=False, click_topics=True
    ):
        """Output a category-ordered g for displaying the main help, grouped by
        category.

        Args:
            cmd_help_dict (dict): A dict `{"category": [topic, topic, ...]}` for
                command-based help.
            db_help_dict (dict): A dict `{"category": [topic, topic], ...]}` for
                database-based help.
            title_lone_category (bool, optional): If a lone category should
                be titled with the category name or not. While pointless in a
                general index, the title should probably show when explicitly
                listing the category itself.
            click_topics (bool, optional): If help-topics are clickable or not
                (for webclient or telnet clients with MXP support).
        Returns:
            str: The help index organized into a grid.

        Notes:
            The input are the pre-loaded help files for commands and database-helpfiles
            respectively. You can override this method to return a custom display of the list of
            commands and topics.

        """

        def _group_by_category(help_dict):
            grid = []
            verbatim_elements = []

            if len(help_dict) == 1 and not title_lone_category:
                # don't list categories if there is only one
                for category in help_dict:
                    # gather and sort the entries from the help dictionary
                    entries = sorted(set(help_dict.get(category, [])))

                    # make the help topics clickable
                    if click_topics:
                        entries = [f"|lchelp {entry}|lt{entry}|le" for entry in entries]

                    # add the entries to the grid
                    grid.extend(entries)
            else:
                # list the categories
                for category in sorted(set(list(help_dict.keys()))):
                    category_str = f"-- {category.title()} "
                    grid.append(
                        ANSIString(
                            self.index_category_clr
                            + category_str
                            + "-" * (width - len(category_str))
                            + self.index_topic_clr
                        )
                    )
                    verbatim_elements.append(len(grid) - 1)

                    # gather and sort the entries from the help dictionary
                    entries = sorted(set(help_dict.get(category, [])))

                    # make the help topics clickable
                    if click_topics:
                        entries = [f"|lchelp {entry}|lt{entry}|le" for entry in entries]

                    # add the entries to the grid
                    grid.extend(entries)

            return grid, verbatim_elements

        help_index = ""
        width = self.client_width()
        grid = []
        verbatim_elements = []
        cmd_grid, db_grid = "", ""

        if any(cmd_help_dict.values()):
            # get the command-help entries by-category
            sep1 = (
                self.index_type_separator_clr
                + pad("Commands", width=width, fillchar="-")
                + self.index_topic_clr
            )
            grid, verbatim_elements = _group_by_category(cmd_help_dict)
            gridrows = format_grid(
                grid,
                width,
                sep="  ",
                verbatim_elements=verbatim_elements,
                line_prefix=self.index_topic_clr,
            )
            cmd_grid = ANSIString("\n").join(gridrows) if gridrows else ""

        if any(db_help_dict.values()):
            # get db-based help entries by-category
            sep2 = (
                self.index_type_separator_clr
                + pad("Game & World", width=width, fillchar="-")
                + self.index_topic_clr
            )
            grid, verbatim_elements = _group_by_category(db_help_dict)
            gridrows = format_grid(
                grid,
                width,
                sep="  ",
                verbatim_elements=verbatim_elements,
                line_prefix=self.index_topic_clr,
            )
            db_grid = ANSIString("\n").join(gridrows) if gridrows else ""

        # only show the main separators if there are actually both cmd and db-based help
        if cmd_grid and db_grid:
            help_index = f"{sep1}\n{cmd_grid}\n{sep2}\n{db_grid}"
        else:
            help_index = f"{cmd_grid}{db_grid}"

        return help_index

    def can_read_topic(self, cmd_or_topic, caller):
        """
        Helper method. If this return True, the given help topic
        be viewable in the help listing. Note that even if this returns False,
        the entry will still be visible in the help index unless `should_list_topic`
        is also returning False.

        Args:
            cmd_or_topic (Command, HelpEntry or FileHelpEntry): The topic/command to test.
            caller: the caller checking for access.

        Returns:
            bool: If command can be viewed or not.

        Notes:
            This uses the 'read' lock. If no 'read' lock is defined, the topic is assumed readable
            by all.

        """
        if inherits_from(cmd_or_topic, "evennia.commands.command.Command"):
            return cmd_or_topic.auto_help and cmd_or_topic.access(caller, "read", default=True)
        else:
            return cmd_or_topic.access(caller, "read", default=True)

    def can_list_topic(self, cmd_or_topic, caller):
        """
        Should the specified command appear in the help table?

        This method only checks whether a specified command should appear in the table of
        topics/commands.  The command can be used by the caller (see the 'should_show_help' method)
        and the command will still be available, for instance, if a character type 'help name of the
        command'.  However, if you return False, the specified command will not appear in the table.
        This is sometimes useful to "hide" commands in the table, but still access them through the
        help system.

        Args:
            cmd_or_topic (Command, HelpEntry or FileHelpEntry): The topic/command to test.
            caller: the caller checking for access.

        Returns:
            bool: If command should be listed or not.

        Notes:
            The `.auto_help` propery is checked for commands. For all help entries,
            the 'view' lock will be checked, and if no such lock is defined, the 'read'
            lock will be used. If neither lock is defined, the help entry is assumed to be
            accessible to all.

        """
        if hasattr(cmd_or_topic, "auto_help") and not cmd_or_topic.auto_help:
            return False

        has_view = (
            "view:" in cmd_or_topic.locks
            if inherits_from(cmd_or_topic, "evennia.commands.command.Command")
            else cmd_or_topic.locks.get("view")
        )

        if has_view:
            return cmd_or_topic.access(caller, "view", default=True)
        else:
            # no explicit 'view' lock - use the 'read' lock
            return cmd_or_topic.access(caller, "read", default=True)

    def collect_topics(self, caller, mode="list"):
        """
        Collect help topics from all sources (cmd/db/file).

        Args:
            caller (Object or Account): The user of the Command.
            mode (str): One of 'list' or 'query', where the first means we are collecting to view
                the help index and the second because of wanting to search for a specific help
                entry/cmd to read. This determines which access should be checked.

        Returns:
            tuple: A tuple of three dicts containing the different types of help entries
            in the order cmd-help, db-help, file-help:
                `({key: cmd,...}, {key: dbentry,...}, {key: fileentry,...}`

        """
        # start with cmd-help
        cmdset = self.cmdset
        # removing doublets in cmdset, caused by cmdhandler
        # having to allow doublet commands to manage exits etc.
        cmdset.make_unique(caller)
        # retrieve all available commands and database / file-help topics.
        # also check the 'cmd:' lock here
        cmd_help_topics = [cmd for cmd in cmdset if cmd and cmd.access(caller, "cmd")]
        # get all file-based help entries, checking perms
        file_help_topics = {topic.key.lower().strip(): topic for topic in FILE_HELP_ENTRIES.all()}
        # get db-based help entries, checking perms
        db_help_topics = {topic.key.lower().strip(): topic for topic in HelpEntry.objects.all()}
        if mode == "list":
            # check the view lock for all help entries/commands and determine key
            cmd_help_topics = {
                cmd.auto_help_display_key if hasattr(cmd, "auto_help_display_key") else cmd.key: cmd
                for cmd in cmd_help_topics
                if self.can_list_topic(cmd, caller)
            }
            db_help_topics = {
                key: entry
                for key, entry in db_help_topics.items()
                if self.can_list_topic(entry, caller)
            }
            file_help_topics = {
                key: entry
                for key, entry in file_help_topics.items()
                if self.can_list_topic(entry, caller)
            }
        else:
            # query - check the read lock on entries
            cmd_help_topics = {
                cmd.auto_help_display_key if hasattr(cmd, "auto_help_display_key") else cmd.key: cmd
                for cmd in cmd_help_topics
                if self.can_read_topic(cmd, caller)
            }
            db_help_topics = {
                key: entry
                for key, entry in db_help_topics.items()
                if self.can_read_topic(entry, caller)
            }
            file_help_topics = {
                key: entry
                for key, entry in file_help_topics.items()
                if self.can_read_topic(entry, caller)
            }

        return cmd_help_topics, db_help_topics, file_help_topics

    def do_search(self, query, entries, search_fields=None):
        """
        Perform a help-query search, default using Lunr search engine.

        Args:
            query (str): The help entry to search for.
            entries (list): All possibilities. A mix of commands, HelpEntries and FileHelpEntries.
            search_fields (list): A list of dicts defining how Lunr will find the
                search data on the elements. If not given, will use a default.

        Returns:
            tuple: A tuple (match, suggestions).

        """
        if not search_fields:
            # lunr search fields/boosts
            search_fields = [
                {"field_name": "key", "boost": 10},
                {"field_name": "aliases", "boost": 7},
                {"field_name": "no_prefix", "boost": 6},
                {"field_name": "category", "boost": 5},
                {"field_name": "tags", "boost": 1},  # tags are not used by default
            ]
        match, suggestions = None, None
        for match_query in (query, f"{query}*"):
            # We first do an exact word-match followed by a start-by query. The
            # return of this will either be a HelpCategory, a Command or a
            # HelpEntry/FileHelpEntry.
            matches, suggestions = help_search_with_index(
                match_query, entries, suggestion_maxnum=self.suggestion_maxnum, fields=search_fields
            )
            if matches:
                match = matches[0]
                break
        return match, suggestions

    def parse(self):
        """
        input is a string containing the command or topic to match.

        The allowed syntax is
        ::

            help <topic>[/<subtopic>[/<subtopic>[/...]]]

        The database/command query is always for `<topic>`, and any subtopics
        is then parsed from there. If a `<topic>` has spaces in it, it is
        always matched before assuming the space begins a subtopic.

        """
        # parse the query

        if self.args:
            self.subtopics = [
                part.strip().lower() for part in self.args.split(self.subtopic_separator_char)
            ]
            self.topic = self.subtopics.pop(0)
        else:
            self.topic = ""
            self.subtopics = []

    def strip_cmd_prefix(self, key, all_keys):
        """
        Conditional strip of a command prefix, such as @ in @desc. By default
        this will be hidden unless there is a duplicate without the prefix
        in the full command set (such as @open and open).

        Args:
            key (str): Command key to analyze.
            all_cmds (list): All command-keys (and potentially aliases).

        Returns:
            str: Potentially modified key to use in help display.

        """
        if key and key[0] in CMD_IGNORE_PREFIXES and key[1:] not in all_keys:
            # filter out e.g. `@` prefixes from display if there is duplicate
            # with the prefix in the set (such as @open/open)
            return key[1:]
        return key

    def func(self):
        """
        Run the dynamic help entry creator.
        """
        caller = self.caller
        query, subtopics, cmdset = self.topic, self.subtopics, self.cmdset
        clickable_topics = self.clickable_topics

        if not query:
            # list all available help entries, grouped by category. We want to
            # build dictionaries {category: [topic, topic, ...], ...}

            cmd_help_topics, db_help_topics, file_help_topics = self.collect_topics(
                caller, mode="list"
            )

            # db-topics override file-based ones
            file_db_help_topics = {**file_help_topics, **db_help_topics}

            # group by category (cmds are listed separately)
            cmd_help_by_category = defaultdict(list)
            file_db_help_by_category = defaultdict(list)

            # get a collection of all keys + aliases to be able to strip prefixes like @
            key_and_aliases = set(chain(*(cmd._keyaliases for cmd in cmd_help_topics.values())))

            for key, cmd in cmd_help_topics.items():
                key = self.strip_cmd_prefix(key, key_and_aliases)
                cmd_help_by_category[cmd.help_category].append(key)
            for key, entry in file_db_help_topics.items():
                file_db_help_by_category[entry.help_category].append(key)

            # generate the index and display
            output = self.format_help_index(
                cmd_help_by_category, file_db_help_by_category, click_topics=clickable_topics
            )
            self.msg_help(output)

            return

        # search for a specific entry. We need to check for 'read' access here before
        # building the set of possibilities.
        cmd_help_topics, db_help_topics, file_help_topics = self.collect_topics(
            caller, mode="query"
        )

        # get a collection of all keys + aliases to be able to strip prefixes like @
        key_and_aliases = set(chain(*(cmd._keyaliases for cmd in cmd_help_topics.values())))

        # db-help topics takes priority over file-help
        file_db_help_topics = {**file_help_topics, **db_help_topics}

        # commands take priority over the other types
        all_topics = {**file_db_help_topics, **cmd_help_topics}

        # get all categories
        all_categories = list(
            set(HelpCategory(topic.help_category) for topic in all_topics.values())
        )

        # all available help options - will be searched in order. We also check # the
        # read-permission here.
        entries = list(all_topics.values()) + all_categories

        # lunr search fields/boosts
        match, suggestions = self.do_search(query, entries)

        if not match:
            # no topic matches found. Only give suggestions.
            help_text = f"There is no help topic matching '{query}'."

            if not suggestions:
                # we don't even have a good suggestion. Run a second search,
                # doing a full-text search in the actual texts of the help
                # entries

                search_fields = [
                    {"field_name": "text", "boost": 1},
                ]

                for match_query in [query, f"{query}*", f"*{query}"]:
                    _, suggestions = help_search_with_index(
                        match_query,
                        entries,
                        suggestion_maxnum=self.suggestion_maxnum,
                        fields=search_fields,
                    )
                    if suggestions:
                        help_text += (
                            "\n... But matches were found within the help "
                            "texts of the suggestions below."
                        )
                        suggestions = [
                            self.strip_cmd_prefix(sugg, key_and_aliases) for sugg in suggestions
                        ]
                        break

            output = self.format_help_entry(
                topic=None,  # this will give a no-match style title
                help_text=help_text,
                suggested=suggestions,
                click_topics=clickable_topics,
            )

            self.msg_help(output)
            return

        if isinstance(match, HelpCategory):
            # no subtopics for categories - these are just lists of topics
            category = match.key
            category_lower = category.lower()
            cmds_in_category = [
                key for key, cmd in cmd_help_topics.items() if category_lower == cmd.help_category
            ]
            topics_in_category = [
                key
                for key, topic in file_db_help_topics.items()
                if category_lower == topic.help_category
            ]
            output = self.format_help_index(
                {category: cmds_in_category},
                {category: topics_in_category},
                title_lone_category=True,
                click_topics=clickable_topics,
            )
            self.msg_help(output)
            return

        if inherits_from(match, "evennia.commands.command.Command"):
            # a command match
            topic = match.key
            help_text = match.get_help(caller, cmdset)
            aliases = match.aliases
            suggested = suggestions[1:]
        else:
            # a database (or file-help) match
            topic = match.key
            help_text = match.entrytext
            aliases = match.aliases if isinstance(match.aliases, list) else match.aliases.all()
            suggested = suggestions[1:]

        # parse for subtopics. The subtopic_map is a dict with the current topic/subtopic
        # text is stored under a `None` key and all other keys are subtopic titles pointing
        # to nested dicts.

        subtopic_map = parse_entry_for_subcategories(help_text)
        help_text = subtopic_map[None]
        subtopic_index = [subtopic for subtopic in subtopic_map if subtopic is not None]

        if subtopics:
            # if we asked for subtopics, parse the found topic_text to see if any match.
            # the subtopics is a list describing the path through the subtopic_map.

            for subtopic_query in subtopics:
                if subtopic_query not in subtopic_map:
                    # exact match failed. Try startswith-match
                    fuzzy_match = False
                    for key in subtopic_map:
                        if key and key.startswith(subtopic_query):
                            subtopic_query = key
                            fuzzy_match = True
                            break

                    if not fuzzy_match:
                        # startswith failed - try an 'in' match
                        for key in subtopic_map:
                            if key and subtopic_query in key:
                                subtopic_query = key
                                fuzzy_match = True
                                break

                    if not fuzzy_match:
                        # no match found - give up
                        checked_topic = topic + f"/{subtopic_query}"
                        output = self.format_help_entry(
                            topic=topic,
                            help_text=f"No help entry found for '{checked_topic}'",
                            subtopics=subtopic_index,
                            click_topics=clickable_topics,
                        )
                        self.msg_help(output)
                        return

                # if we get here we have an exact or fuzzy match

                subtopic_map = subtopic_map.pop(subtopic_query)
                subtopic_index = [subtopic for subtopic in subtopic_map if subtopic is not None]
                # keep stepping down into the tree, append path to show position
                topic = topic + f"/{subtopic_query}"

            # we reached the bottom of the topic tree
            help_text = subtopic_map[None]

        topic = self.strip_cmd_prefix(topic, key_and_aliases)
        if subtopics:
            aliases = None
        else:
            aliases = [self.strip_cmd_prefix(alias, key_and_aliases) for alias in aliases]
        suggested = [self.strip_cmd_prefix(sugg, key_and_aliases) for sugg in suggested]

        output = self.format_help_entry(
            topic=topic,
            help_text=help_text,
            aliases=aliases,
            subtopics=subtopic_index,
            suggested=suggested,
            click_topics=clickable_topics,
        )

        self.msg_help(output)


def _loadhelp(caller):
    entry = caller.db._editing_help
    if entry:
        return entry.entrytext
    else:
        return ""


def _savehelp(caller, buffer):
    entry = caller.db._editing_help
    caller.msg("Saved help entry.")
    if entry:
        entry.entrytext = buffer


def _quithelp(caller):
    caller.msg("Closing the editor.")
    del caller.db._editing_help


class CmdSetHelp(CmdHelp):
    """
    Edit the help database.

    Usage:
      sethelp[/switches] <topic>[[;alias;alias][,category[,locks]] [= <text>]

    Switches:
      edit - open a line editor to edit the topic's help text.
      replace - overwrite existing help topic.
      append - add text to the end of existing topic with a newline between.
      extend - as append, but don't add a newline.
      delete - remove help topic.

    Examples:
      sethelp lore = In the beginning was ...
      sethelp/append pickpocketing,Thievery = This steals ...
      sethelp/replace pickpocketing, ,attr(is_thief) = This steals ...
      sethelp/edit thievery

    If not assigning a category, the `settings.DEFAULT_HELP_CATEGORY` category
    will be used. If no lockstring is specified, everyone will be able to read
    the help entry.  Sub-topics are embedded in the help text.

    Note that this cannot modify command-help entries - these are modified
    in-code, outside the game.

    # SUBTOPICS

    ## Adding subtopics

    Subtopics helps to break up a long help entry into sub-sections. Users can
    access subtopics with |whelp topic/subtopic/...|n Subtopics are created and
    stored together with the main topic.

    To start adding subtopics, add the text '# SUBTOPICS' on a new line at the
    end of your help text. After this you can now add any number of subtopics,
    each starting with '## <subtopic-name>' on a line, followed by the
    help-text of that subtopic.
    Use '### <subsub-name>' to add a sub-subtopic and so on. Max depth is 5. A
    subtopic's title is case-insensitive and can consist of multiple words -
    the user will be able to enter a partial match to access it.

    For example:

    | Main help text for <topic>
    |
    | # SUBTOPICS
    |
    | ## about
    |
    | Text for the '<topic>/about' subtopic'
    |
    | ### more about-info
    |
    | Text for the '<topic>/about/more about-info sub-subtopic
    |
    | ## extra
    |
    | Text for the '<topic>/extra' subtopic

    """

    key = "sethelp"
    aliases = []
    switch_options = ("edit", "replace", "append", "extend", "delete")
    locks = "cmd:perm(Helper)"
    help_category = "Building"
    arg_regex = None

    def parse(self):
        """We want to use the default parser rather than the CmdHelp.parse"""
        return COMMAND_DEFAULT_CLASS.parse(self)

    def func(self):
        """Implement the function"""

        switches = self.switches
        lhslist = self.lhslist

        if not self.args:
            self.msg(
                "Usage: sethelp[/switches] <topic>[;alias;alias][,category[,locks,..] = <text>"
            )
            return

        nlist = len(lhslist)
        topicstr = lhslist[0] if nlist > 0 else ""
        if not topicstr:
            self.msg("You have to define a topic!")
            return
        topicstrlist = topicstr.split(";")
        topicstr, aliases = (
            topicstrlist[0],
            topicstrlist[1:] if len(topicstr) > 1 else [],
        )
        aliastxt = ("(aliases: %s)" % ", ".join(aliases)) if aliases else ""
        old_entry = None

        # check if we have an old entry with the same name

        cmd_help_topics, db_help_topics, file_help_topics = self.collect_topics(
            self.caller, mode="query"
        )
        # db-help topics takes priority over file-help
        file_db_help_topics = {**file_help_topics, **db_help_topics}
        # commands take priority over the other types
        all_topics = {**file_db_help_topics, **cmd_help_topics}
        # get all categories
        all_categories = list(
            set(HelpCategory(topic.help_category) for topic in all_topics.values())
        )
        # all available help options - will be searched in order. We also check # the
        # read-permission here.
        entries = list(all_topics.values()) + all_categories

        # default setup
        category = lhslist[1] if nlist > 1 else DEFAULT_HELP_CATEGORY
        lockstring = ",".join(lhslist[2:]) if nlist > 2 else "read:all()"

        # search for existing entries of this or other types
        old_entry = None
        for querystr in topicstrlist:
            match, _ = self.do_search(querystr, entries)
            if match:
                warning = None
                if isinstance(match, HelpCategory):
                    warning = (
                        f"'{querystr}' matches (or partially matches) the name of "
                        f"help-category '{match.key}'. If you continue, your help entry will "
                        "take precedence and the category (or part of its name) *may* not "
                        "be usable for grouping help entries anymore."
                    )
                elif inherits_from(match, "evennia.commands.command.Command"):
                    warning = (
                        f"'{querystr}' matches (or partially matches) the key/alias of "
                        f"Command '{match.key}'. Command-help take precedence over other "
                        "help entries so your help *may* be impossible to reach for those "
                        "with access to that command."
                    )
                elif inherits_from(match, "evennia.help.filehelp.FileHelpEntry"):
                    warning = (
                        f"'{querystr}' matches (or partially matches) the name/alias of the "
                        f"file-based help topic '{match.key}'. File-help entries cannot be "
                        "modified from in-game (they are files on-disk). If you continue, "
                        "your help entry may shadow the file-based one's name partly or "
                        "completely."
                    )
                if warning:
                    # show a warning for a clashing help-entry type. Even if user accepts this
                    # we don't break here since we may need to show warnings for other inputs.
                    # We don't count this as an old-entry hit because we can't edit these
                    # types of entries.
                    self.msg(f"|rWarning:\n|r{warning}|n")
                    repl = yield ("|wDo you still want to continue? Y/[N]?|n")
                    if repl.lower() not in ("y", "yes"):
                        self.msg("Aborted.")
                        return
                else:
                    # a db-based help entry - this is OK
                    old_entry = match
                    category = lhslist[1] if nlist > 1 else old_entry.help_category
                    lockstring = ",".join(lhslist[2:]) if nlist > 2 else old_entry.locks.get()
                    break

        category = category.lower()

        if "edit" in switches:
            # open the line editor to edit the helptext. No = is needed.
            if old_entry:
                topicstr = old_entry.key
                if self.rhs:
                    # we assume append here.
                    old_entry.entrytext += "\n%s" % self.rhs
                helpentry = old_entry
            else:
                helpentry = create.create_help_entry(
                    topicstr,
                    self.rhs,
                    category=category,
                    locks=lockstring,
                    aliases=aliases,
                )
            self.caller.db._editing_help = helpentry

            EvEditor(
                self.caller,
                loadfunc=_loadhelp,
                savefunc=_savehelp,
                quitfunc=_quithelp,
                key="topic {}".format(topicstr),
                persistent=True,
            )
            return

        if "append" in switches or "merge" in switches or "extend" in switches:
            # merge/append operations
            if not old_entry:
                self.msg(f"Could not find topic '{topicstr}'. You must give an exact name.")
                return
            if not self.rhs:
                self.msg("You must supply text to append/merge.")
                return
            if "merge" in switches:
                old_entry.entrytext += " " + self.rhs
            else:
                old_entry.entrytext += "\n%s" % self.rhs
            old_entry.aliases.add(aliases)
            self.msg(f"Entry updated:\n{old_entry.entrytext}{aliastxt}")
            return

        if "delete" in switches or "del" in switches:
            # delete the help entry
            if not old_entry:
                self.msg(f"Could not find topic '{topicstr}'{aliastxt}.")
                return
            old_entry.delete()
            self.msg(f"Deleted help entry '{topicstr}'{aliastxt}.")
            return

        # at this point it means we want to add a new help entry.
        if not self.rhs:
            self.msg("You must supply a help text to add.")
            return
        if old_entry:
            if "replace" in switches:
                # overwrite old entry
                old_entry.key = topicstr
                old_entry.entrytext = self.rhs
                old_entry.help_category = category
                old_entry.locks.clear()
                old_entry.locks.add(lockstring)
                old_entry.aliases.add(aliases)
                old_entry.save()
                self.msg(f"Overwrote the old topic '{topicstr}'{aliastxt}.")
            else:
                self.msg(
                    f"Topic '{topicstr}'{aliastxt} already exists. Use /edit to open in editor, or "
                    "/replace, /append and /merge to modify it directly."
                )
        else:
            # no old entry. Create a new one.
            new_entry = create.create_help_entry(
                topicstr, self.rhs, category=category, locks=lockstring, aliases=aliases
            )
            if new_entry:
                self.msg(f"Topic '{topicstr}'{aliastxt} was successfully created.")
                if "edit" in switches:
                    # open the line editor to edit the helptext
                    self.caller.db._editing_help = new_entry
                    EvEditor(
                        self.caller,
                        loadfunc=_loadhelp,
                        savefunc=_savehelp,
                        quitfunc=_quithelp,
                        key="topic {}".format(new_entry.key),
                        persistent=True,
                    )
                    return
            else:
                self.msg(f"Error when creating topic '{topicstr}'{aliastxt}! Contact an admin.")



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-07-06 20:41


from django.db import connection, migrations


def _table_exists(db_cursor, tablename):
    "Returns bool if table exists or not"
    return tablename in connection.introspection.table_names()


class Migration(migrations.Migration):

    dependencies = [("comms", "0014_auto_20170705_1736")]

    db_cursor = connection.cursor()

    if not _table_exists(db_cursor, "channels.channeldb_db_receivers_players"):
        # OBS - this is run BEFORE migrations are run!
        operations = []
    else:
        operations = [
            migrations.RemoveField(
                model_name="channeldb",
                name="db_subscriptions",  # this is now db_account_subscriptions
            ),
            migrations.RemoveField(model_name="msg", name="db_receivers_players"),
            migrations.RemoveField(model_name="msg", name="db_sender_players"),
        ]



"""
Components - ChrisLR 2022

This file contains the base class to inherit for creating new components.
"""

from evennia.commands.cmdset import CmdSet

from . import exceptions
from .listing import COMPONENT_LISTING


class BaseComponent(type):
    """
    This is the metaclass for components,
    responsible for registering components to the listing.
    """

    @classmethod
    def __new__(cls, *args):
        """
        Every class that uses this metaclass will be registered
        as a component in the Component Listing using its name.
        All of them require a unique name.
        """
        new_type = super().__new__(*args)
        if new_type.__base__ == object:
            return new_type

        name = getattr(new_type, "name", None)
        if not name:
            raise ValueError(f"Component {new_type} requires a name.")

        if existing_type := COMPONENT_LISTING.get(name):
            if not str(new_type) == str(existing_type):
                raise ValueError(f"Component name {name} is a duplicate, must be unique.")
        else:
            COMPONENT_LISTING[name] = new_type

        return new_type


class Component(metaclass=BaseComponent):
    """
    This is the base class for components.
    Any component must inherit from this class to be considered for usage.

    Each Component must supply the name, it is used as a slot name but also part of the attribute key.
    """

    __slots__ = ("host",)

    name = ""
    slot = None

    _fields = {}

    def __init__(self, host=None):
        assert self.name, "All Components must have a name"
        self.host = host

    @classmethod
    def default_create(cls, host):
        """
        This is called when the host is created
         and should return the base initialized state of a component.

        Args:
            host (object): The host typeclass instance

        Returns:
            Component: The created instance of the component

        """
        new = cls(host)
        return new

    @classmethod
    def create(cls, host, **kwargs):
        """
        This is the method to call when supplying kwargs to initialize a component.

        Args:
            host (object): The host typeclass instance
            **kwargs: Key-Value of default values to replace.
                      To persist the value, the key must correspond to a DBField.

        Returns:
            Component: The created instance of the component

        """

        new = cls.default_create(host)
        for key, value in kwargs.items():
            setattr(new, key, value)

        return new

    def cleanup(self):
        """
        This deletes all component attributes from the host's db
        """
        for name in self._fields.keys():
            delattr(self, name)

    @classmethod
    def load(cls, host):
        """
        Loads a component instance
         This is called whenever a component is loaded (ex: Server Restart)

        Args:
            host (object): The host typeclass instance

        Returns:
            Component: The loaded instance of the component

        """
        return cls(host)

    def at_added(self, host):
        """
        This is the method called when a component is registered on a host.

        Args:
            host (object): The host typeclass instance

        """
        if self.host and self.host != host:
            raise exceptions.InvalidComponentError("Components must not register twice!")

        self.host = host

    def at_removed(self, host):
        """
        This is the method called when a component is removed from a host.

        Args:
            host (object): The host typeclass instance

        """
        if host != self.host:
            raise ValueError("Component attempted to remove from the wrong host.")

        self.host = None

    @property
    def attributes(self):
        """
        Shortcut property returning the host's AttributeHandler.

        Returns:
            AttributeHandler: The Host's AttributeHandler

        """
        return self.host.attributes

    @property
    def nattributes(self):
        """
        Shortcut property returning the host's In-Memory AttributeHandler (Non Persisted).

        Returns:
            AttributeHandler: The Host's In-Memory AttributeHandler

        """
        return self.host.nattributes

    @classmethod
    def add_field(cls, name, field):
        cls._fields[name] = field

    @classmethod
    def get_fields(cls):
        return tuple(cls._fields.values())

    @classmethod
    def get_component_slot(cls):
        return cls.slot or cls.name



"""
XYZGrid - Griatch 2021

"""
from . import (
    example,
    launchcmd,
    prototypes,
    tests,
    utils,
    xymap,
    xymap_legend,
    xyzgrid,
    xyzroom,
)

from . import commands  # isort:skip  - this needs to be imported last



"""
Base objects for the Evscaperoom contrib.


The object class itself provide the actions possible to use on that object.
This makes these objects suitable for use with multi-inheritance. For example,
to make an object both possible to smell and eat or drink, find the appropriate
parents in this module and make an object like this:

class Apple(Edible, Smellable):

    def at_drink(self, caller):
        # ...

    def at_smell(self, caller):
        # ...

Various object parents could be more complex, so read the class for more info.

Available parents:

- EvscapeRoomObject - parent class for all Evscaperoom entities (also the room itself)
- Feelable
- Listenable
- Smellable
- Rotatable
- Openable
- Readable
- IndexReadable  (like a lexicon you have to give a search term in)
- Movable
- Edible
- Drinkable
- Usable
- Insertable  (can be inserted into a target)
- Combinable  (combines with another object to create a new one)
- Mixable     (used for mixing potions into it)
- HasButtons  (an object with buttons on it)
- CodeInput   (code locks)
- Sittable    (can be sat on)
- Liable      (can be lied down on)
- Kneeable    (can be kneed down on)
- Climbable   (can be climbed on)
- Positionable (supports sit/lie/knee/climb at once)

"""
import inspect
import re

from evennia import DefaultObject
from evennia.utils.utils import list_to_string, wrap

from .utils import create_evscaperoom_object, parse_for_perspectives, parse_for_things


class EvscaperoomObject(DefaultObject):
    """
    Default object base for all objects related to the contrib.

    """

    # these will be automatically filtered out by self.parse for
    # focus-commands using arguments like (`combine [with] object`)
    # override this per-class as necessary.
    action_prepositions = ("in", "with", "on", "into", "to")

    # this mapping allows for prettier descriptions of our current
    # position
    position_prep_map = {"sit": "sitting", "kneel": "kneeling", "lie": "lying", "climb": "standing"}

    def at_object_creation(self):
        """
        Called once when object is first created.

        """
        # state flags (setup/reset for each state).
        self.db.tagcategory = None
        self.db.flags = {}

        self.db.desc = "Nothing of interest."

        self.db.positions = {}

    _tagcategory = None

    @property
    def tagcategory(self):
        if not self._tagcategory:
            self._tagcategory = (
                self.location.db.tagcategory if self.location else self.db.tagcategory
            )
        return self._tagcategory

    @property
    def room(self):
        return self.location or self

    @property
    def roomstate(self):
        return self.room.statehandler.current_state

    def next_state(self, statename=None):
        """
        Helper to have the object switch the room to next state

        Args:
            statename (str, optional): If given, move to this
                state next. Otherwise use the default next-state
                of the current state.

        """
        self.room.statehandler.next_state(next_state=statename)

    def set_flag(self, flagname):
        "Set flag on object"
        self.db.flags[flagname] = True

    def unset_flag(self, flagname):
        "Unset flag on object"
        if flagname in self.db.flags:
            del self.db.flags[flagname]

    def check_flag(self, flagname):
        "Check if flag is set on this object"
        return self.db.flags.get(flagname, False)

    def set_character_flag(self, char, flagname, value=True):
        "Set flag on character"
        flags = char.attributes.get(flagname, category=self.tagcategory, default={})
        flags[flagname] = value
        char.attributes.add(flagname, flags, category=self.tagcategory)

    def unset_character_flag(self, char, flagname):
        "Set flag on character"
        flags = char.attributes.get(flagname, category=self.tagcategory, default={})
        if flagname in flags:
            flags.pop(flagname, None)
            char.attributes.add(flagname, flags, category=self.tagcategory)

    def check_character_flag(self, char, flagname):
        "Check if flag is set on character"
        flags = char.attributes.get(flagname, category=self.tagcategory, default={})
        return flags.get(flagname, False)

    def msg_room(self, caller, string, skip_caller=False):
        """
        Message everyone in the room with a message that is parsed for
        ~first/third person grammar, as well as for *thing markers.

        Args:
            caller (Object or None): Sender of the message. If None, there
                is no sender.
            string (str): Message to parse and send to the room.
            skip_caller (bool): Send to everyone except caller.

        Notes:
            Messages sent by this method will be tagged with a type of
            'your_action' and `others_action`. This is an experiment for
            allowing users of e.g. the webclient to redirect messages to
            differnt windows.

        """
        you = caller.key if caller else "they"
        first_person, third_person = parse_for_perspectives(string, you=you)
        for char in self.room.get_all_characters():
            options = char.attributes.get("options", category=self.room.tagcategory, default={})
            style = options.get("things_style", 2)
            if char == caller:
                if not skip_caller:
                    txt = parse_for_things(first_person, things_style=style)
                    char.msg((txt, {"type": "your_action"}))
            else:
                txt = parse_for_things(third_person, things_style=style)
                char.msg((txt, {"type": "others_action"}))

    def msg_char(self, caller, string, client_type="your_action"):
        """
        Send message only to caller (not to the room at large)

        """
        # we must clean away markers
        first_person, _ = parse_for_perspectives(string)
        options = caller.attributes.get("options", category=self.room.tagcategory, default={})
        style = options.get("things_style", 2)
        txt = parse_for_things(first_person, things_style=style)
        caller.msg((txt, {"type": client_type}))

    def msg_system(self, message, target=None, borders=True):
        """
        Send a 'system message' by using the State.msg function.
        """
        self.room.state.msg(message, target=target, borders=borders)

    def get_position(self, caller):
        """
        Get position of caller on this object (like lying, sitting, kneeling,
        standing). See the Positionable child class.

        Args:
            caller (Object): The one position we seek.

        Returns:
            obj, pos (Object, str): The object we have a position relative to,
                as well as the name of that position (lying, sitting, kneeling,
                standing).  If these are None, it means we are standing on the
                floor.

        """
        pos = caller.attributes.get("position", category=self.tagcategory)
        if pos:
            obj, old_position = pos
            return obj, old_position
        return None, None

    def set_position(self, caller, new_position):
        """
        Set position of caller (like lying, sitting, kneeling, standing)
        on this object. See Positionable child class.

        Args:
            caller (Object): The one positioning themselves on this object.
            new_position (str, None): One of "lie", "kneel", "sit" or "stand".
                If `None`, remove position (character stands normally on the
                floor).

        """
        if new_position is None:
            # reset position
            caller.attributes.remove("position", category=self.tagcategory)
            if caller in self.db.positions:
                del self.db.positions[caller]
        else:
            # set a new position on this object
            position = (self, new_position)
            caller.attributes.add("position", position, category=self.tagcategory)
            self.db.positions[caller] = new_position

    def at_focus(self, caller):
        """
        Called when somone is focusing on this object.

        Args:
            caller (Character): The one doing the focusing.

        """
        self.msg_char(caller, caller.at_look(self), client_type="look")

    def at_unfocus(self, caller):
        """
        Called when focus leaves this object. Note that more than one caller
        may be focusing on the object at the same time, so we should not change
        the state of the object itself here!

        Args:
            caller (Character): The one doing the unfocusing.

        """
        pass

    def at_speech(self, speaker, action):
        """
        We don't use the default at_say hook since we handle the send logic in
        the command. This is only meant to trigger eventual game-events when
        speaking to an object or the room.

        Args:
            speaker (Character): The one speaking.
            action (str): One of 'say', 'whisper' or 'shout'

        """
        pass

    def parse(self, args):
        """
        Simple parser of focus arguments starting with a preposition,
        like 'combine with <object>' <- we want to strip out the preposition
        here.

        """
        args = re.sub(
            r"|".join(r"^{}\s".format(prep) for prep in self.action_prepositions), "", args
        )
        return args

    def get_cmd_signatures(self):
        """
        This allows the object to return more detailed call signs
        for each of their at_focus_* methods. This is useful for
        things like detailed arguments (only 'move' but 'move left/right')

        Returns:
            callsigns (list, None): List of strings to inject into the
                available action list produced by `self.get_help`. If `None`,
                automatically find actions based on the method names.
            custom_helpstr (str): This should be the help text for
                the command with a marker `{callsigns}` for where to
                inject the list of callsigns.

        """
        command_signatures = []
        helpstr = ""
        methods = inspect.getmembers(self, predicate=inspect.ismethod)
        for name, method in methods:
            if name.startswith("at_focus_"):
                command_signatures.append(name[9:])
        command_signatures = sorted(command_signatures)

        if len(command_signatures) == 1:
            helpstr = f"It looks like {self.key} may be " "suitable to {callsigns}."
        else:
            helpstr = (
                f"At first glance, it looks like {self.key} might be " "suitable to {callsigns}."
            )
        return command_signatures, helpstr

    def get_short_desc(self, full_desc):
        """
        Extract the first sentence from the desc and use as the short desc.

        """
        mat = re.match(r"(^.*?[.?!])", full_desc.strip(), re.M + re.U + re.I + re.S)
        if mat:
            return mat.group(0).strip()
        return full_desc

    def get_help(self, caller):
        """
        Get help about this object. By default we return a
        listing of all actions you can do on this object.

        """
        # custom-created signatures. We don't sort these
        command_signatures, helpstr = self.get_cmd_signatures()

        callsigns = list_to_string(["*" + sig for sig in command_signatures], endsep="or")

        # parse for *thing markers (use these as items)
        options = caller.attributes.get("options", category=self.room.tagcategory, default={})
        style = options.get("things_style", 2)

        helpstr = helpstr.format(callsigns=callsigns)
        helpstr = parse_for_things(helpstr, style, clr="|w")
        return wrap(helpstr, width=80)

    # Evennia hooks

    def return_appearance(self, looker, **kwargs):
        """Could be modified per state. We generally don't worry about the
        contents of the object by default.

        """
        # accept a custom desc
        desc = kwargs.get("desc", self.db.desc)

        if kwargs.get("unfocused", False):
            # use the shorter description
            focused = ""
            desc = self.get_short_desc(desc)
            helptxt = ""
        else:
            focused = " |g(examining |G- use '|gex|G' again to look away. See also '|ghelp|G')|n"
            helptxt = kwargs.get("helptxt", f"\n\n({self.get_help(looker)})")

        obj, pos = self.get_position(looker)
        pos = (
            f" |w({self.position_prep_map[pos]} on " f"{obj.get_display_name(looker)})"
            if obj
            else ""
        )

        return f" ~~ |y{self.get_display_name(looker)}|n{focused}{pos}|n ~~\n\n{desc}{helptxt}"


class Feelable(EvscaperoomObject):
    """
    Any object that you can feel the surface of.

    """

    def at_focus_feel(self, caller, **kwargs):
        self.msg_char(caller, f"You feel *{self.key}.")


class Listenable(EvscaperoomObject):
    """
    Any object one can listen to.

    """

    def at_focus_listen(self, caller, **kwargs):
        self.msg_char(caller, f"You listen to *{self.key}")


class Smellable(EvscaperoomObject):
    """
    Any object you can smell.

    """

    def at_focus_smell(self, caller, **kwargs):
        self.msg_char(caller, f"You smell *{self.key}.")


class Rotatable(EvscaperoomObject):
    """
    Any object that you can lift up and look at from different angles

    """

    rotate_flag = "rotatable"
    start_rotatable = True

    def at_object_creation(self):
        super().at_object_creation()

        if self.start_rotatable:
            self.set_flag("rotatable")

    def at_focus_rotate(self, caller, **kwargs):
        if self.check_flag("rotatable"):
            self.at_rotate(caller)
        else:
            self.at_cannot_rotate(caller)

    at_focus_turn = at_focus_rotate

    def at_rotate(self, caller):
        self.msg_char(caller, f"You turn *{self.key} around.")

    def at_cannot_rotate(self, caller):
        self.msg_char(caller, f"You cannot rotate this.")


class Openable(EvscaperoomObject):
    """
    Any object that you can open/close. It's lockable with
    a flag.

    """

    # this flag must be set for item to open. None for unlocked.
    unlock_flag = "unlocked"
    open_flag = "open"
    # start this item in the opened/unlocked state
    start_open = False

    def at_object_creation(self):
        super().at_object_creation()
        if self.start_open:
            self.set_flag(self.unlock_flag)
            self.set_flag(self.open_flag)

    def at_focus_open(self, caller, **kwargs):
        if self.check_flag(self.open_flag):
            self.at_already_open(caller)
        elif self.unlock_flag is None or self.check_flag(self.unlock_flag):
            self.set_flag(self.open_flag)
            self.at_open(caller)
        else:
            self.at_locked(caller)

    def at_focus_close(self, caller, **kwargs):
        if not self.check_flag(self.open_flag):
            self.at_already_closed(caller)
        else:
            self.unset_flag(self.open_flag)
            self.at_close(caller)

    def at_open(self, caller):
        self.msg_char(caller, f"You open *{self.key}")

    def at_already_open(self, caller):
        self.msg_char(caller, f"{self.key.capitalize()} is already open.")

    def at_locked(self, caller):
        self.msg_char(caller, f"{self.key.capitalize()} won't open.")

    def at_close(self, caller):
        self.msg_char(caller, f"You close *{self.key}.")

    def at_already_closed(self, caller):
        self.msg_char(caller, f"{self.key.capitalize()} is already closed.")


class Readable(EvscaperoomObject):
    """
    Any object that you can read from. This is controlled
    from a flag.

    """

    # this must be set to be able to read. None to
    # always be able to read.

    read_flag = "readable"
    start_readable = True

    def at_object_creation(self):
        super().at_object_creation()
        if self.start_readable:
            self.set_flag(self.read_flag)

    def at_focus_read(self, caller, **kwargs):
        if self.read_flag is None or self.check_flag(self.read_flag):
            self.at_read(caller)
        else:
            self.at_cannot_read(caller)

    def at_read(self, caller, *args, **kwargs):
        self.msg_char(caller, f"You read from *{self.key}.")

    def at_cannot_read(self, caller, *args, **kwargs):
        self.msg_char(caller, "You cannot understand a thing!")


class IndexReadable(Readable):
    """
    Any object for which you need to specify a key/index to get a given result
    back. For example a lexicon or book where you enter a topic or a page
    number to see what's to be read on that page.
    """

    # keys should be lower-key
    index = {"page1": "This is page1", "page2": "This is page2", "page two": "page2"}  # alias

    def at_focus_read(self, caller, **kwargs):
        topic = kwargs.get("args").strip().lower()

        entry = self.index.get(topic, None)

        if entry is None or not self.check_flag(self.read_flag):
            self.at_cannot_read(caller, topic)
        else:
            if entry in self.index:
                # an alias-reroute
                entry = self.index[entry]
            self.at_read(caller, topic, entry)

    def get_cmd_signatures(self):
        txt = (
            f"You don't have the time to read this from beginning to end. "
            "Use *read <topic> to look up something in particular."
        )
        return [], txt

    def at_cannot_read(self, caller, topic, *args, **kwargs):
        self.msg_char(caller, f"Cannot find an entry on '{topic}'.")

    def at_read(self, caller, topic, entry, *args, **kwargs):
        self.msg_char(caller, f"You read about '{topic}':\n{entry.strip()}")


class Movable(EvscaperoomObject):
    """
    Any object that can be moved from one place to another
    or in one direction or another.

    Once moved to a given position, the object's state will
    change.

    """

    # these are the possible locations (or directions) to move to
    # name: callable
    move_positions = {"left": "at_left", "right": "at_right"}
    start_position = "left"

    def at_object_creation(self):
        super().at_object_creation()
        self.db.position = self.start_position

    def get_cmd_signatures(self):
        txt = "Looks like you can {callsigns}."
        return ["move", "push", "shove left/right"], txt

    def at_focus_move(self, caller, **kwargs):
        pos = self.parse(kwargs["args"])
        callfunc_name = self.move_positions.get(pos)

        if callfunc_name:
            if self.db.position == pos:
                self.at_already_moved(caller, pos)
            else:
                self.db.position = pos
                getattr(self, callfunc_name)(caller)
        else:
            self.at_cannot_move(caller)

    at_focus_shove = at_focus_move
    at_focus_push = at_focus_move

    def at_cannot_move(self, caller):
        self.msg_char(caller, "That does not work.")

    def at_already_moved(self, caller, position):
        self.msg_char(caller, f"You already moved *{self.key} to the {position}.")

    def at_left(self, caller):
        self.msg_char(caller, f"You move *{self.key} left")

    def at_right(self, caller):
        self.msg_char(caller, f"You move *{self.key} right")


class BaseConsumable(EvscaperoomObject):
    """
    Any object that is consumable in some way. This acts as an
    abstract parent.

    This sets a flag that
    is unique for each person consuming, allowing it to e.g. only
    be consumed once (don't support multi-uses here, that's left for
    a custom object if needed).

    """

    consume_flag = "consume"
    # may only consume once
    one_consume_only = True

    def handle_consume(self, caller, action, **kwargs):
        """
        Wrap this by the at_focus method
        """
        if self.one_consume_only and self.has_consumed(caller):
            self.at_already_consumed(caller, action)
        else:
            self.has_consumed(caller, True)
            self.at_consume(caller, action)

    def has_consumed(self, caller, setflag=False):
        "Check if caller already consumed at least once"
        flag = f"{self.consume_flag}#{caller.id}"
        if setflag:
            self.set_flag(flag)
        else:
            return self.check_flag(flag)

    def at_consume(self, caller, action):
        if hasattr(self, f"at_{action}"):
            getattr(self, f"at_{action}")(caller)
        else:
            self.msg_char(caller, f"You {action} *{self.key}.")

    def at_already_consumed(self, caller, action):
        self.msg_char(caller, f"You can't {action} any more.")


class Edible(BaseConsumable):
    """
    Any object specifically possible to eat.

    """

    consume_flag = "eat"

    def at_focus_eat(self, caller, **kwargs):
        super().handle_consume(caller, "eat", **kwargs)


class Drinkable(BaseConsumable):
    """
    Any object specifically possible to drink.

    """

    consume_flag = "drink"

    def at_focus_drink(self, caller, **kwargs):
        super().handle_consume(caller, "drink", **kwargs)

    def at_focus_sip(self, caller, **kwargs):
        super().handle_consume(caller, "sip", **kwargs)

    def at_consume(self, caller, action):
        self.msg_char(caller, f"You {action} from *{self.key}.")

    def at_already_consumed(self, caller, action):
        self.msg_char(caller, f"You can't drink any more.")


class BaseApplicable(EvscaperoomObject):
    """
    Any object that can be applied/inserted/used on another object in some way.
    This acts an an abstract base class.

    """

    # the target object this is to be used with must
    # have this flag. It'll likely be unique to this
    # object combination.
    target_flag = "applicable"

    def handle_apply(self, caller, action, **kwargs):
        """
        Wrap this with the at_focus methods in the child classes

        """
        args = self.parse(kwargs["args"])
        if not args:
            self.msg_char(caller, "You need to specify a target.")
            return
        obj = caller.search(args)
        if not obj:
            return
        try:
            can_apply = obj.check_flag(self.target_flag)
        except AttributeError:
            can_apply = False
        if can_apply:
            self.at_apply(caller, action, obj)
        else:
            self.at_cannot_apply(caller, action, obj)

    def at_apply(self, caller, action, obj):
        self.msg_char(caller, f"You {action} *{self.key} to {obj.key}.")

    def at_cannot_apply(self, caller, action, obj):
        self.msg_char(caller, f"You cannot {action} *{self.key} to {obj.key}.")


class Usable(BaseApplicable):
    """
    Any object that can be used with another object.

    """

    target_flag = "usable"

    def at_focus_use(self, caller, **kwargs):
        super().handle_apply(caller, "use", **kwargs)

    def at_apply(self, caller, action, obj):
        self.msg_char(caller, f"You {action} *{self.key} with {obj.key}")

    def at_cannot_apply(self, caller, action, obj):
        self.msg_char(caller, f"You cannot {action} *{self.key} with {obj.key}.")


class Insertable(BaseApplicable):
    """
    Any object that can be inserted into another object.

    This would cover a key, for example.

    """

    # this would likely be a custom name
    target_flag = "insertable"

    def at_focus_insert(self, caller, **kwargs):
        super().handle_apply(caller, "insert", **kwargs)

    def at_apply(self, caller, action, obj):
        self.msg_char(caller, f"You {action} *{self.key} in {obj.key}.")

    def get_cmd_signatures(self):
        txt = "You can use this object to {callsigns}"
        return ["insert in <object>"], txt

    def at_cannot_apply(self, caller, action, obj):
        self.msg_char(caller, f"You cannot {action} *{self.key} in {obj.key}.")


class Combinable(BaseApplicable):
    """
    Any object that combines with another object to create
    a new one.

    """

    # the other object must have this flag to be able to be combined
    # (this is likely unique for a given combination)
    target_flag = "combinable"
    # create-dict to pass into the create_object for the
    # new "combined" object.
    new_create_dict = {
        "typeclass": "evscaperoom.objects.Combinable",
        "key": "sword",
        "aliases": ["combined"],
    }
    # if set, destroy the two components used to make the new one
    destroy_components = True

    def at_focus_combine(self, caller, **kwargs):
        super().at_focus_apply(caller, **kwargs)

    def get_cmd_signatures(self):
        txt = "It looks like this should work: {callsigns}"
        return ["combine <object>"], txt

    def at_cannot_apply(self, caller, action, obj):
        self.msg_char(caller, f"You cannot {action} *{self.key} with {obj.key}.")

    def at_apply(self, caller, action, other_obj):
        create_dict = self.new_create_dict
        if "location" not in create_dict:
            create_dict["location"] = self.location
        new_obj = create_evscaperoom_object(**create_dict)
        if new_obj and self.destroy_components:
            self.msg_char(
                caller, f"You combine *{self.key} with {other_obj.key} to make {new_obj.key}!"
            )
            other_obj.delete()
            self.delete()


class Mixable(EvscaperoomObject):
    """
    Any object into which you can mix ingredients (such as when
    mixing a potion). This offers no actions on its own, instead
    the ingredients should be 'used' with this object in order
    mix, calling at_mix when they do.
    """

    # ingredients can check for this before they allow to mix at all
    mixer_flag = "mixer"
    # ingredients must have these flags and this order
    ingredient_recipe = ["ingredient1", "ingredient2", "ingredient3"]

    def at_object_creation(self):
        super().at_object_creation()
        self.set_flag(self.mixer_flag)
        # this holds the ingredients as they are added
        self.db.ingredients = []

    def check_mixture(self):
        "check so mixture is correct, returning True/False."
        ingredients = list(self.db.ingredients)
        for iflag, flag in enumerate(self.ingredient_recipe):
            try:
                if not ingredients[iflag].check_flag(flag):
                    return False
            except (IndexError, AttributeError):
                return False
        # we only get here if all ingredients have the right flags in the right
        # order
        return True

    def handle_mix(self, caller, ingredient, **kwargs):
        """
        Add ingredient object to mixture.

        Called by the mixing ingredient. We assume the ingredient has already
        checked to make sure they allow themselves to be mixed into an object
        with this mixer_flag.

        """
        self.db.ingredients.append(ingredient)
        # normal mix
        self.at_mix(caller, ingredient, **kwargs)

        if len(self.db.ingredients) >= len(self.ingredient_recipe):
            # we have enough, check if it matches recipe

            if self.check_mixture():
                self.at_mix_success(caller, ingredient, **kwargs)
            else:
                self.room.log(
                    f"{self.name} mix failure: Tried {' + '.join([ing.key for ing in self.db.ingredients if ing])}"
                )
                self.db.ingredients = []
                self.at_mix_failure(caller, ingredient, **kwargs)

    def at_mix(self, caller, ingredient, **kwargs):
        self.msg_room(caller, f"~You ~mix {ingredient.key} into *{self.key}.")

    def at_mix_failure(self, caller, ingredient, **kwargs):
        self.msg_room(caller, f"This mix doesn't work. ~You ~clean and start over.")

    def at_mix_success(self, caller, ingredient, **kwargs):
        self.msg_room(caller, f"~You successfully ~complete the mix!")


class HasButtons(EvscaperoomObject):
    """
    Any object with buttons to push/press

    """

    # mapping keys/aliases to calling method
    buttons = {
        "green button": "at_green_button",
        "green": "at_green_button",
        "red button": "at_red_button",
        "red": "at_red_button",
    }

    def get_cmd_signatures(self):
        helptxt = (
            "It looks like you should be able to operate "
            f"*{self.key} by means of "
            "{callsigns}."
        )
        return ["push", "press red/green button"], helptxt

    def at_focus_press(self, caller, **kwargs):
        arg = self.parse(kwargs["args"])
        callfunc_name = self.buttons.get(arg)
        if callfunc_name:
            getattr(self, callfunc_name)(caller)
        else:
            self.at_nomatch(caller)

    at_focus_push = at_focus_press

    def at_nomatch(self, caller):
        self.msg_char(caller, "That does not seem right.")

    def at_green_button(self, caller):
        self.msg_char(caller, "You press the green button.")

    def at_red_button(self, caller):
        self.msg_char(caller, "You press the red button.")


class CodeInput(EvscaperoomObject):
    """
    Any object where you can enter a code of some sort
    to have an effect happen.

    """

    # the code of this
    code = "PASSWORD"
    code_hint = "eight letters A-Z"
    case_insensitive = True
    # code locked no matter what is input
    infinitely_locked = False

    def at_focus_code(self, caller, **kwargs):
        args = self.parse(kwargs["args"].strip())

        if not args:
            self.at_no_code(caller)
            return
        if self.infinitely_locked:
            code_correct = False
        elif self.case_insensitive:
            code_correct = args.upper() == self.code.upper()
        else:
            code_correct = args == self.code

        if code_correct:
            self.at_code_correct(caller, args)
        else:
            self.at_code_incorrect(caller, args)

    def get_cmd_signatures(self):
        helptxt = "Looks like you need to use {callsigns}."
        return ["code <code>"], helptxt

    def at_no_code(self, caller):
        self.msg_char(caller, f"Looks like you need to enter |w{self.code_hint}|n.")

    def at_code_correct(self, caller, code_tried):
        self.msg_char(caller, "That's the right code!")

    def at_code_incorrect(self, caller, code_tried):
        self.msg_char(caller, f"That's not the right code (need {self.code_hint}).")


class BasePositionable(EvscaperoomObject):
    """
    Any object a character can be positioned on. This is meant as an
    abstract parent.

    This is a little special since a char can only have one position at a
    time and must therefore be aware of the other 'positional' actions
    any object may support (otherwise you may end up sitting/standing/etc on
    more than one object at once!)

    We set a Attribute (obj, position) on the caller to indicate that
    they have a position on an object. This is necessary so as to not have
    the caller sit on more than one sittable object at a time, for example. The
    'positions' Attribute on this object holds a mapping of who is sitting
    lying etc on this object.  We don't add a limit to how many chars could
    have a position on an object - it's not realistic, but this goes with the
    philosophy that one character should not be able to block others if they go
    inactive etc.

    This state is also tied to the general 'stand' command, which should return
    the player to the normal standing state regardless of if they focus on this
    object or not.

    """

    def at_object_creation(self):
        super().at_object_creation()
        # mapping {object: position}.
        self.db.positions = {}

    def handle_position(self, caller, new_pos, **kwargs):
        """
        Wrap this with the at_focus_ method of the child class.

        """
        old_obj, old_pos = self.get_position(caller)
        if old_obj:
            if old_obj is self:
                if old_pos == new_pos:
                    self.at_again_position(caller, new_pos)
                else:
                    self.set_position(caller, new_pos)
                    self.at_position(caller, new_pos)
            else:
                self.at_cannot_position(caller, new_pos, old_obj, old_pos)
        else:
            self.set_position(caller, new_pos)
            self.at_position(caller, new_pos)

    def at_cannot_position(self, caller, position, old_obj, old_pos):
        self.msg_char(
            caller,
            f"You can't; you are currently {self.position_prep_map[old_pos]} on *{old_obj.key} "
            "(better |wstand|n first).",
        )

    def at_again_position(self, caller, position):
        self.msg_char(
            caller, f"But you are already {self.position_prep_map[position]} on *{self.key}?"
        )

    def at_position(self, caller, position):
        self.msg_room(caller, f"~You ~{position} on *{self.key}.")


class Sittable(BasePositionable):
    """
    Any object you can sit on.

    """

    def at_focus_sit(self, caller, **kwargs):
        super().handle_position(caller, "sit", **kwargs)


class Liable(BasePositionable):
    """
    Any object you can lie down on.

    """

    def at_focus_lie(self, caller, **kwargs):
        super().handle_position(caller, "lie", **kwargs)


class Kneelable(BasePositionable):
    """
    Any object you can kneel on.

    """

    def at_focus_kneel(self, caller, **kwargs):
        super().handle_position(caller, "kneel", **kwargs)


class Climbable(BasePositionable):
    """
    Any object you can climb up to stand on. We name this
    'climb' so as to not collide with the general 'stand'
    command, which resets your position.

    """

    def at_focus_climb(self, caller, **kwargs):
        super().handle_position(caller, "climb", **kwargs)


class Positionable(Sittable, Liable, Kneelable, Climbable):
    """
    An object on which you can position yourself in one of the
    supported ways (sit, lie, kneel or climb)

    """

    def get_cmd_signatures(self):
        txt = "It looks like you can {callsigns} on it."
        return ["sit", "lie", "kneel", "climb"], txt



"""
Evadventure Techdemo area

This is is used for testing specific features without any story or overall gameplay

While this looks like a Python module, it is meant to run with the `batchcode` processor in-game:

    batchcode evadventure.build_techdemo

This will step through the #CODE blocks of this file to build the tech demo. For help with using
the batchcode processor, see the
[processor documentation](https://www.evennia.com/docs/latest/Components/Batch-Code-Processor.html).

You can also build/rebuild individiaul #CODE blocks in the `batchcode/interactive` mode.

"""

# HEADER

# this is loaded at the top of every #CODE block

from evennia import DefaultExit, create_object, search_object
from evennia.contrib.tutorials.evadventure import npcs
from evennia.contrib.tutorials.evadventure.dungeon import (
    EvAdventureDungeonStartRoom,
    EvAdventureDungeonStartRoomExit,
)
from evennia.contrib.tutorials.evadventure.objects import EvAdventureWeapon
from evennia.contrib.tutorials.evadventure.rooms import (
    EvAdventurePvPRoom,
    EvAdventureRoom,
)

# CODE

# Hub room evtechdemo#00
# for other test areas to link back to. Connects in turn back to Limbo.

limbo = search_object("Limbo")[0]
hub_room = create_object(
    EvAdventureRoom,
    key="Techdemo Hub",
    aliases=("evtechdemo#00",),
    attributes=[("desc", "Central hub for EvAdventure tech demo.")],
)
create_object(
    DefaultExit,
    key="EvAdventure Techdemo",
    aliases=("techdemo", "demo", "evadventure"),
    location=limbo,
    destination=hub_room,
)
create_object(
    DefaultExit,
    key="Back to Limbo",
    aliases=("limbo", "back"),
    location=hub_room,
    destination=limbo,
)


# CODE

# A combat room evtechdemo#01
# with a static enemy

combat_room = create_object(EvAdventurePvPRoom, key="Combat Arena", aliases=("evtechdemo#01",))
# link to/back to/from hub
hub_room = search_object("evtechdemo#00")[0]
create_object(
    DefaultExit, key="combat test", aliases=("combat",), location=hub_room, destination=combat_room
)
create_object(
    DefaultExit,
    key="Back to Hub",
    aliases=("back", "hub"),
    location=combat_room,
    destination=hub_room,
)
# create training dummy with a stick
combat_room_enemy = create_object(
    npcs.EvAdventureMob, key="Training Dummy", aliases=("dummy",), location=combat_room
)
weapon_stick = create_object(EvAdventureWeapon, key="stick", attributes=(("damage_roll", "1d2"),))
combat_room_enemy.weapon = weapon_stick


# CODE

# A dungeon start room for testing the dynamic dungeon generation.

dungeon_start_room = create_object(
    EvAdventureDungeonStartRoom,
    key="Dungeon start room",
    aliases=("evtechdemo#02",),
    attributes=(("desc", "A central room, with dark exits leading to mysterious fates."),),
)
# link to/back to/from hub
hub_room = search_object("evtechdemo#00")[0]
create_object(
    DefaultExit,
    key="dungeon test",
    aliases=("dungeon",),
    location=hub_room,
    destination=dungeon_start_room,
)
create_object(
    DefaultExit,
    key="Back to Hub",
    aliases=("back", "hub"),
    location=dungeon_start_room,
    destination=hub_room,
)

# add special exits out of the dungeon start room.
# These must have one of the 8 cardinal directions
# we point these exits back to the same location, which
# is what the system will use to trigger generating a new room
create_object(
    EvAdventureDungeonStartRoomExit,
    key="north",
    aliases=("n",),
    location=dungeon_start_room,
    destination=dungeon_start_room,
)
create_object(
    EvAdventureDungeonStartRoomExit,
    key="east",
    aliases=("e",),
    location=dungeon_start_room,
    destination=dungeon_start_room,
)
create_object(
    EvAdventureDungeonStartRoomExit,
    key="south",
    aliases=("s",),
    location=dungeon_start_room,
    destination=dungeon_start_room,
)
create_object(
    EvAdventureDungeonStartRoomExit,
    key="west",
    aliases=("w",),
    location=dungeon_start_room,
    destination=dungeon_start_room,
)



# -*- coding: utf-8 -*-


from django.db import migrations, models


def remove_manage_scripts(apps, schema_editor):
    ScriptDB = apps.get_model("scripts", "ScriptDB")
    for script in ScriptDB.objects.filter(
        db_typeclass_path__in=(
            "src.scripts.scripts.CheckSessions",
            "src.scripts.scripts.ValidateScripts",
            "src.scripts.scripts.ValidateChannelHandler",
            "src.scripts.scripts.ValidateIdmapperCache",
            "src.utils.gametime.GameTime",
        )
    ):
        script.delete()


class Migration(migrations.Migration):

    dependencies = [("scripts", "0005_auto_20150306_1441")]

    operations = [migrations.RunPython(remove_manage_scripts)]



"""
Web plugin hooks.
"""


def at_webserver_root_creation(web_root):
    """
    This is called as the web server has finished building its default
    path tree. At this point, the media/ and static/ URIs have already
    been added to the web root.

    Args:
        web_root (twisted.web.resource.Resource): The root
            resource of the URI tree. Use .putChild() to
            add new subdomains to the tree.

    Returns:
        web_root (twisted.web.resource.Resource): The potentially
            modified root structure.

    Example:
        from twisted.web import static
        my_page = static.File("web/mypage/")
        my_page.indexNames = ["index.html"]
        web_root.putChild("mypage", my_page)

    """
    return web_root


def at_webproxy_root_creation(web_root):
    """
    This function can modify the portal proxy service.
    Args:
        web_root (evennia.server.webserver.Website): The Evennia
            Website application. Use .putChild() to add new
            subdomains that are Portal-accessible over TCP;
            primarily for new protocol development, but suitable
            for other shenanigans.
    """
    return web_root



"""
Components - ChrisLR 2022

This file contains the classes that allow a typeclass to use components.
"""

from . import exceptions, signals
from .component import Component
from .listing import get_component_class


class ComponentProperty:
    """
    This allows you to register a component on a typeclass.
    Components registered with this property are automatically added
    to any instance of this typeclass.

    Defaults can be overridden for this typeclass by passing kwargs
    """

    def __init__(self, name, **kwargs):
        """
        Initializes the descriptor

        Args:
            name (str): The name of the component
            **kwargs (any): Key=Values overriding default values of the component
        """
        self.name = name
        self.values = kwargs
        self.component_class = None
        self.slot_name = None

    def __get__(self, instance, owner):
        if not self.component_class:
            component_class = get_component_class(self.name)
            self.component_class = component_class
            self.slot_name = component_class.get_component_slot()

        component = instance.components.get(self.slot_name)
        return component

    def __set__(self, instance, value):
        raise Exception("Cannot set a class property")

    def __set_name__(self, owner, name):
        # Retrieve the class_components set on the direct class only
        class_components = owner.__dict__.get("_class_components", [])
        if not class_components:
            setattr(owner, "_class_components", class_components)

        class_components.append((self.name, self.values))


class ComponentHandler:
    """
    This is the handler that will be added to any typeclass that inherits from ComponentHolder.
    It lets you add or remove components and will load components as needed.
    It stores the list of registered components on the host .db with component_names as key.
    """

    def __init__(self, host):
        self.host = host
        self._loaded_components = {}

    def add(self, component: Component):
        """
        Method to add a Component to a host.
        It caches the loaded component and appends its name to the host's component name list.
        It will also call the component's 'at_added' method, passing its host.

        Args:
            component (object): The 'loaded' component instance to add.

        """
        component_name = component.name
        self.db_names.append(component_name)
        self.host.tags.add(component_name, category="components")
        self._set_component(component)
        for field in component.get_fields():
            field.at_added(component)

        component.at_added(self.host)

    def add_default(self, name):
        """
        Method to add a Component initialized to default values on a host.
        It will retrieve the proper component and instantiate it with 'default_create'.
        It will cache this new component and add it to its list.
        It will also call the component's 'at_added' method, passing its host.

        Args:
            name (str): The name of the component class to add.

        """
        component_class = get_component_class(name)
        component_instance = component_class.default_create(self.host)
        self.add(component_instance)

    def remove(self, component: Component):
        """
        Method to remove a component instance from a host.
        It removes the component from the cache and listing.
        It will call the component's 'at_removed' method.

        Args:
            component (object): The component instance to remove.

        """
        name = component.name
        slot_name = component.get_component_slot()
        if not self.has(slot_name):
            message = f"Cannot remove {name} from {self.host.name} as it is not registered."
            raise exceptions.ComponentIsNotRegistered(message)

        for field in component.get_fields():
            field.at_removed(component)

        component.at_removed(self.host)

        self.host.tags.remove(component.name, category="components")
        self.host.signals.remove_object_listeners_and_responders(component)

        self.db_names.remove(name)
        del self._loaded_components[slot_name]

    def remove_by_name(self, name):
        """
        Method to remove a component instance from a host.
        It removes the component from the cache and listing.
        It will call the component's 'at_removed' method.

        Args:
            name (str): The name of the component to remove or its slot.

        """
        instance = self.get(name)
        if not instance:
            message = f"Cannot remove {name} from {self.host.name} as it is not registered."
            raise exceptions.ComponentIsNotRegistered(message)

        self.remove(instance)

    def get(self, name: str) -> Component | None:
        return self._loaded_components.get(name)

    def has(self, name: str) -> bool:
        """
        Method to check if a component is registered and ready.

        Args:
            name (str): The name of the component or the slot.

        """
        return name in self._loaded_components

    def initialize(self):
        """
        Method that loads and caches each component currently registered on the host.
        It retrieves the names from the registered listing and calls 'load' on each
        prototype class that can be found from this listing.

        """
        component_names = self.db_names
        if not component_names:
            return

        for component_name in component_names:
            component = get_component_class(component_name)
            if component:
                component_instance = component.load(self.host)
                self._set_component(component_instance)
            else:
                message = (
                    f"Could not initialize runtime component {component_name} of {self.host.name}"
                )
                raise exceptions.ComponentDoesNotExist(message)

    def _set_component(self, component):
        """
        Sets the loaded component in this instance.
        """
        slot_name = component.get_component_slot()
        self._loaded_components[slot_name] = component
        self.host.signals.add_object_listeners_and_responders(component)

    @property
    def db_names(self):
        """
        Property shortcut to retrieve the registered component keys

        Returns:
            component_names (iterable): The name of each component that is registered

        """
        names = self.host.attributes.get("component_names")
        if names is None:
            self.host.db.component_names = []
            names = self.host.db.component_names

        return names

    def __getattr__(self, name):
        return self.get(name)


class ComponentHolderMixin:
    """
    Mixin to add component support to a typeclass

    Components are set on objects using the component.name as an object attribute.
    All registered components are initialized on the typeclass.
    They will be of None value if not present in the class components or runtime components.
    """

    def at_init(self):
        """
        Method that initializes the ComponentHandler.
        """
        super(ComponentHolderMixin, self).at_init()
        setattr(self, "_component_handler", ComponentHandler(self))
        setattr(self, "_signal_handler", signals.SignalsHandler(self))
        self.components.initialize()
        self.signals.trigger("at_after_init")

    def at_post_puppet(self, *args, **kwargs):
        super().at_post_puppet(*args, **kwargs)
        self.signals.trigger("at_post_puppet", *args, **kwargs)

    def at_post_unpuppet(self, *args, **kwargs):
        super().at_post_unpuppet(*args, **kwargs)
        self.signals.trigger("at_post_unpuppet", *args, **kwargs)

    def basetype_setup(self):
        """
        Method that initializes the ComponentHandler, creates and registers all
        components that were set on the typeclass using ComponentProperty.
        """
        super().basetype_setup()
        setattr(self, "_component_handler", ComponentHandler(self))
        setattr(self, "_signal_handler", signals.SignalsHandler(self))
        class_components = self._get_class_components()
        for component_name, values in class_components:
            component_class = get_component_class(component_name)
            component = component_class.create(self, **values)
            self.components.add(component)

        self.signals.trigger("at_basetype_setup")

    @property
    def components(self) -> ComponentHandler:
        """
        Property getter to retrieve the component_handler.
        Returns:
            ComponentHandler: This Host's ComponentHandler
        """
        return getattr(self, "_component_handler", None)

    @property
    def cmp(self) -> ComponentHandler:
        """
        Shortcut Property getter to retrieve the component_handler.
        Returns:
            ComponentHandler: This Host's ComponentHandler
        """
        return self.components

    @property
    def signals(self) -> signals.SignalsHandler:
        return getattr(self, "_signal_handler", None)

    def _get_class_components(self):
        class_components = {}

        def base_type_iterator():
            base_stack = [type(self)]
            while base_stack:
                _base_type = base_stack.pop()
                yield _base_type
                base_stack.extend(_base_type.__bases__)

        for base_type in base_type_iterator():
            base_class_components = getattr(base_type, "_class_components", ())
            for cmp_name, cmp_values in base_class_components:
                cmp_class = get_component_class(cmp_name)
                cmp_slot = cmp_class.get_component_slot()
                class_components[cmp_slot] = (cmp_name, cmp_values)

        instance_components = getattr(self, "_class_components", ())
        for cmp_name, cmp_values in instance_components:
            cmp_class = get_component_class(cmp_name)
            cmp_slot = cmp_class.get_component_slot()
            class_components[cmp_slot] = (cmp_name, cmp_values)

        return tuple(class_components.values())



"""
This is a convenient container gathering all the main
search methods for the various database tables.

It is intended to be used e.g. as

> from evennia.utils import search
> match = search.objects(...)

Note that this is not intended to be a complete listing of all search
methods! You need to refer to the respective manager to get all
possible search methods. To get to the managers from your code, import
the database model and call its 'objects' property.

Also remember that all commands in this file return lists (also if
there is only one match) unless noted otherwise.

Example: To reach the search method 'get_object_with_account'
         in evennia/objects/managers.py:

> from evennia.objects.models import ObjectDB
> match = Object.objects.get_object_with_account(...)


"""

# Import the manager methods to be wrapped

from django.contrib.contenttypes.models import ContentType
from django.db.utils import OperationalError, ProgrammingError

# limit symbol import from API
__all__ = (
    "search_object",
    "search_account",
    "search_script",
    "search_message",
    "search_channel",
    "search_help_entry",
    "search_tag",
    "search_script_tag",
    "search_account_tag",
    "search_channel_tag",
    "search_typeclass",
)


# import objects this way to avoid circular import problems
try:
    ObjectDB = ContentType.objects.get(app_label="objects", model="objectdb").model_class()
    AccountDB = ContentType.objects.get(app_label="accounts", model="accountdb").model_class()
    ScriptDB = ContentType.objects.get(app_label="scripts", model="scriptdb").model_class()
    Msg = ContentType.objects.get(app_label="comms", model="msg").model_class()
    ChannelDB = ContentType.objects.get(app_label="comms", model="channeldb").model_class()
    HelpEntry = ContentType.objects.get(app_label="help", model="helpentry").model_class()
    Tag = ContentType.objects.get(app_label="typeclasses", model="tag").model_class()
except (OperationalError, ProgrammingError):
    # this is a fallback used during tests/doc building
    print("Database not available yet - using temporary fallback for search managers.")
    from evennia.accounts.models import AccountDB
    from evennia.comms.models import ChannelDB, Msg
    from evennia.help.models import HelpEntry
    from evennia.objects.models import ObjectDB
    from evennia.scripts.models import ScriptDB
    from evennia.typeclasses.tags import Tag  # noqa

# -------------------------------------------------------------------
# Search manager-wrappers
# -------------------------------------------------------------------

#
# Search objects as a character
#
# NOTE: A more powerful wrapper of this method
#  is reachable from within each command class
#  by using self.caller.search()!
#
#    def object_search(self, ostring=None,
#                      attribute_name=None,
#                      typeclass=None,
#                      candidates=None,
#                      exact=True):
#
#        Search globally or in a list of candidates and return results.
#        The result is always a list of Objects (or the empty list)
#
#        Arguments:
#        ostring: (str) The string to compare names against. By default (if
#                  not attribute_name is set), this will search object.key
#                  and object.aliases in order. Can also be on the form #dbref,
#                  which will, if exact=True be matched against primary key.
#        attribute_name: (str): Use this named ObjectAttribute to match ostring
#                        against, instead of the defaults.
#        typeclass (str or TypeClass): restrict matches to objects having
#                  this typeclass. This will help speed up global searches.
#        candidates (list obj ObjectDBs): If supplied, search will only be
#                  performed among the candidates in this list. A common list
#                  of candidates is the contents of the current location.
#        exact (bool): Match names/aliases exactly or partially. Partial
#                  matching matches the beginning of words in the names/aliases,
#                  using a matching routine to separate multiple matches in
#                  names with multiple components (so "bi sw" will match
#                  "Big sword"). Since this is more expensive than exact
#                  matching, it is recommended to be used together with
#                  the objlist keyword to limit the number of possibilities.
#                  This keyword has no meaning if attribute_name is set.
#
#        Returns:
#        A list of matching objects (or a list with one unique match)
#    def object_search(self, ostring, caller=None,
#                      candidates=None,
#                      attribute_name=None):
#
search_object = ObjectDB.objects.search_object
search_objects = search_object
object_search = search_object
objects = search_objects

#
# Search for accounts
#
# account_search(self, ostring)

#     Searches for a particular account by name or
#     database id.
#
#     ostring = a string or database id.
#

search_account = AccountDB.objects.search_account
search_accounts = search_account
account_search = search_account
accounts = search_accounts

#
#   Searching for scripts
#
# script_search(self, ostring, obj=None, only_timed=False)
#
#     Search for a particular script.
#
#     ostring - search criterion - a script ID or key
#     obj - limit search to scripts defined on this object
#     only_timed - limit search only to scripts that run
#                  on a timer.
#

search_script = ScriptDB.objects.search_script
search_scripts = search_script
script_search = search_script
scripts = search_scripts
#
# Searching for communication messages
#
#
# message_search(self, sender=None, receiver=None, channel=None, freetext=None)
#
#     Search the message database for particular messages. At least one
#     of the arguments must be given to do a search.
#
#     sender - get messages sent by a particular account
#     receiver - get messages received by a certain account
#     channel - get messages sent to a particular channel
#     freetext - Search for a text string in a message.
#                NOTE: This can potentially be slow, so make sure to supply
#                one of the other arguments to limit the search.
#

search_message = Msg.objects.search_message
search_messages = search_message
message_search = search_message
messages = search_messages

#
# Search for Communication Channels
#
# channel_search(self, ostring)
#
#     Search the channel database for a particular channel.
#
#     ostring - the key or database id of the channel.
#     exact -  requires an exact ostring match (not case sensitive)
#

search_channel = ChannelDB.objects.search_channel
search_channels = search_channel
channel_search = search_channel
channels = search_channels

#
# Find help entry objects.
#
# search_help(self, ostring, help_category=None)
#
#     Retrieve a search entry object.
#
#     ostring - the help topic to look for
#     category - limit the search to a particular help topic
#

search_help = HelpEntry.objects.search_help
search_help_entry = search_help
search_help_entries = search_help
help_entry_search = search_help
help_entries = search_help


# Locate Attributes

#    search_object_attribute(key, category, value, strvalue) (also search_attribute works)
#    search_account_attribute(key, category, value, strvalue) (also search_attribute works)
#    search_script_attribute(key, category, value, strvalue) (also search_attribute works)
#    search_channel_attribute(key, category, value, strvalue) (also search_attribute works)

# Note that these return the object attached to the Attribute,
# not the attribute object itself (this is usually what you want)


def search_object_attribute(
    key=None, category=None, value=None, strvalue=None, attrtype=None, **kwargs
):
    return ObjectDB.objects.get_by_attribute(
        key=key, category=category, value=value, strvalue=strvalue, attrtype=attrtype, **kwargs
    )


def search_account_attribute(
    key=None, category=None, value=None, strvalue=None, attrtype=None, **kwargs
):
    return AccountDB.objects.get_by_attribute(
        key=key, category=category, value=value, strvalue=strvalue, attrtype=attrtype, **kwargs
    )


def search_script_attribute(
    key=None, category=None, value=None, strvalue=None, attrtype=None, **kwargs
):
    return ScriptDB.objects.get_by_attribute(
        key=key, category=category, value=value, strvalue=strvalue, attrtype=attrtype, **kwargs
    )


def search_channel_attribute(
    key=None, category=None, value=None, strvalue=None, attrtype=None, **kwargs
):
    return ChannelDB.objects.get_by_attribute(
        key=key, category=category, value=value, strvalue=strvalue, attrtype=attrtype, **kwargs
    )


# search for attribute objects
search_attribute_object = ObjectDB.objects.get_attribute

# Locate Tags

#    search_object_tag(key=None, category=None) (also search_tag works)
#    search_account_tag(key=None, category=None)
#    search_script_tag(key=None, category=None)
#    search_channel_tag(key=None, category=None)

# Note that this returns the object attached to the tag, not the tag
# object itself (this is usually what you want)


def search_object_by_tag(key=None, category=None, tagtype=None, **kwargs):
    """
    Find object based on tag or category.

    Args:
        key (str, optional): The tag key to search for.
        category (str, optional): The category of tag
            to search for. If not set, uncategorized
            tags will be searched.
        tagtype (str, optional): 'type' of Tag, by default
            this is either `None` (a normal Tag), `alias` or
            `permission`. This always apply to all queried tags.
        kwargs (any): Other optional parameter that may be supported
            by the manager method.

    Returns:
        matches (list): List of Objects with tags matching
            the search criteria, or an empty list if no
            matches were found.

    """
    return ObjectDB.objects.get_by_tag(key=key, category=category, tagtype=tagtype, **kwargs)


search_tag = search_object_by_tag  # this is the most common case


def search_account_tag(key=None, category=None, tagtype=None, **kwargs):
    """
    Find account based on tag or category.

    Args:
        key (str, optional): The tag key to search for.
        category (str, optional): The category of tag
            to search for. If not set, uncategorized
            tags will be searched.
        tagtype (str, optional): 'type' of Tag, by default
            this is either `None` (a normal Tag), `alias` or
            `permission`. This always apply to all queried tags.
        kwargs (any): Other optional parameter that may be supported
            by the manager method.

    Returns:
        matches (list): List of Accounts with tags matching
            the search criteria, or an empty list if no
            matches were found.

    """
    return AccountDB.objects.get_by_tag(key=key, category=category, tagtype=tagtype, **kwargs)


def search_script_tag(key=None, category=None, tagtype=None, **kwargs):
    """
    Find script based on tag or category.

    Args:
        key (str, optional): The tag key to search for.
        category (str, optional): The category of tag
            to search for. If not set, uncategorized
            tags will be searched.
        tagtype (str, optional): 'type' of Tag, by default
            this is either `None` (a normal Tag), `alias` or
            `permission`. This always apply to all queried tags.
        kwargs (any): Other optional parameter that may be supported
            by the manager method.

    Returns:
        matches (list): List of Scripts with tags matching
            the search criteria, or an empty list if no
            matches were found.

    """
    return ScriptDB.objects.get_by_tag(key=key, category=category, tagtype=tagtype, **kwargs)


def search_channel_tag(key=None, category=None, tagtype=None, **kwargs):
    """
    Find channel based on tag or category.

    Args:
        key (str, optional): The tag key to search for.
        category (str, optional): The category of tag
            to search for. If not set, uncategorized
            tags will be searched.
        tagtype (str, optional): 'type' of Tag, by default
            this is either `None` (a normal Tag), `alias` or
            `permission`. This always apply to all queried tags.
        kwargs (any): Other optional parameter that may be supported
            by the manager method.

    Returns:
        matches (list): List of Channels with tags matching
            the search criteria, or an empty list if no
            matches were found.

    """
    return ChannelDB.objects.get_by_tag(key=key, category=category, tagtype=tagtype, **kwargs)


# search for tag objects (not the objects they are attached to
search_tag_object = ObjectDB.objects.get_tag


# Locate Objects by Typeclass

# search_objects_by_typeclass(typeclass="", include_children=False, include_parents=False) (also search_typeclass works)
# This returns the objects of the given typeclass


def search_objects_by_typeclass(typeclass, include_children=False, include_parents=False):
    """
    Searches through all objects returning those of a certain typeclass.

    Args:
        typeclass (str or class): A typeclass class or a python path to a typeclass.
        include_children (bool, optional): Return objects with
            given typeclass *and* all children inheriting from this
            typeclass. Mutuall exclusive to `include_parents`.
        include_parents (bool, optional): Return objects with
            given typeclass *and* all parents to this typeclass.
            Mutually exclusive to `include_children`.

    Returns:
        objects (list): The objects found with the given typeclasses.
    """
    return ObjectDB.objects.typeclass_search(
        typeclass=typeclass,
        include_children=include_children,
        include_parents=include_parents,
    )


search_typeclass = search_objects_by_typeclass



"""
Test Color markup.

"""

import re

from evennia.utils.test_resources import BaseEvenniaTest

from . import color_markups


class TestColorMarkup(BaseEvenniaTest):
    """
    Note: Normally this would be tested by importing the ansi parser and run
    the mappings through it. This is not possible since the ansi module creates
    its mapping at the module/class level; since the ansi module is used by so
    many other modules it appears that trying to overload
    settings to test it causes issues with unrelated tests.
    """

    def test_curly_markup(self):
        ansi_map = color_markups.CURLY_COLOR_ANSI_EXTRA_MAP
        self.assertIsNotNone(re.match(re.escape(ansi_map[7][0]), "{r"))
        self.assertIsNotNone(re.match(re.escape(ansi_map[-1][0]), "{[X"))
        xterm_fg = color_markups.CURLY_COLOR_XTERM256_EXTRA_FG
        self.assertIsNotNone(re.match(xterm_fg[0], "{001"))
        self.assertIsNotNone(re.match(xterm_fg[0], "{123"))
        self.assertIsNotNone(re.match(xterm_fg[0], "{455"))
        xterm_bg = color_markups.CURLY_COLOR_XTERM256_EXTRA_BG
        self.assertIsNotNone(re.match(xterm_bg[0], "{[001"))
        self.assertIsNotNone(re.match(xterm_bg[0], "{[123"))
        self.assertIsNotNone(re.match(xterm_bg[0], "{[455"))
        xterm_gfg = color_markups.CURLY_COLOR_XTERM256_EXTRA_GFG
        self.assertIsNotNone(re.match(xterm_gfg[0], "{=h"))
        self.assertIsNotNone(re.match(xterm_gfg[0], "{=e"))
        self.assertIsNotNone(re.match(xterm_gfg[0], "{=w"))
        xterm_gbg = color_markups.CURLY_COLOR_XTERM256_EXTRA_GBG
        self.assertIsNotNone(re.match(xterm_gbg[0], "{[=a"))
        self.assertIsNotNone(re.match(xterm_gbg[0], "{[=k"))
        self.assertIsNotNone(re.match(xterm_gbg[0], "{[=z"))
        bright_map = color_markups.CURLY_COLOR_ANSI_XTERM256_BRIGHT_BG_EXTRA_MAP
        self.assertEqual(bright_map[0][1], "{[500")
        self.assertEqual(bright_map[-1][1], "{[222")

    def test_mux_markup(self):
        ansi_map = color_markups.MUX_COLOR_ANSI_EXTRA_MAP
        self.assertIsNotNone(re.match(re.escape(ansi_map[10][0]), "%cr"))
        self.assertIsNotNone(re.match(re.escape(ansi_map[-1][0]), "%cX"))
        xterm_fg = color_markups.MUX_COLOR_XTERM256_EXTRA_FG
        self.assertIsNotNone(re.match(xterm_fg[0], "%c001"))
        self.assertIsNotNone(re.match(xterm_fg[0], "%c123"))
        self.assertIsNotNone(re.match(xterm_fg[0], "%c455"))
        xterm_bg = color_markups.MUX_COLOR_XTERM256_EXTRA_BG
        self.assertIsNotNone(re.match(xterm_bg[0], "%c[001"))
        self.assertIsNotNone(re.match(xterm_bg[0], "%c[123"))
        self.assertIsNotNone(re.match(xterm_bg[0], "%c[455"))
        xterm_gfg = color_markups.MUX_COLOR_XTERM256_EXTRA_GFG
        self.assertIsNotNone(re.match(xterm_gfg[0], "%c=h"))
        self.assertIsNotNone(re.match(xterm_gfg[0], "%c=e"))
        self.assertIsNotNone(re.match(xterm_gfg[0], "%c=w"))
        xterm_gbg = color_markups.MUX_COLOR_XTERM256_EXTRA_GBG
        self.assertIsNotNone(re.match(xterm_gbg[0], "%c[=a"))
        self.assertIsNotNone(re.match(xterm_gbg[0], "%c[=k"))
        self.assertIsNotNone(re.match(xterm_gbg[0], "%c[=z"))
        bright_map = color_markups.MUX_COLOR_ANSI_XTERM256_BRIGHT_BG_EXTRA_MAP
        self.assertEqual(bright_map[0][1], "%c[500")
        self.assertEqual(bright_map[-1][1], "%c[222")



"""
MUD ruleset based on the _Knave_ OSR tabletop RPG by Ben Milton (modified for MUD use).

The center of the rule system is the "RollEngine", which handles all rolling of dice
and determining what the outcome is.

----

"""
from random import randint

from .enums import Ability
from .random_tables import death_and_dismemberment as death_table

# Basic rolls


class EvAdventureRollEngine:
    """
    This groups all dice rolls of EvAdventure. These could all have been normal functions, but we
    are group them in a class to make them easier to partially override and replace later.

    """

    def roll(self, roll_string, max_number=10):
        """
        NOTE: In evennia/contribs/rpg/dice/ is a more powerful dice roller with
        more features, such as modifiers, secret rolls etc. This is much simpler and only
        gets a simple sum of normal rpg-dice.

        Args:
            roll_string (str): A roll using standard rpg syntax, <number>d<diesize>, like
                1d6, 2d10 etc. Max die-size is 1000.
            max_number (int): The max number of dice to roll. Defaults to 10, which is usually
                more than enough.

        Returns:
            int: The rolled result - sum of all dice rolled.

        Raises:
            TypeError: If roll_string is not on the right format or otherwise doesn't validate.

        Notes:
            Since we may see user input to this function, we make sure to validate the inputs (we
            wouldn't bother much with that if it was just for developer use).

        """
        max_diesize = 1000
        roll_string = roll_string.lower()
        if "d" not in roll_string:
            raise TypeError(
                f"Dice roll '{roll_string}' was not recognized. Must be `<number>d<dicesize>`."
            )
        number, diesize = roll_string.split("d", 1)
        try:
            number = int(number)
            diesize = int(diesize)
        except Exception:
            raise TypeError(f"The number and dice-size of '{roll_string}' must be numerical.")
        if 0 < number > max_number:
            raise TypeError(f"Invalid number of dice rolled (must be between 1 and {max_number})")
        if 0 < diesize > max_diesize:
            raise TypeError(f"Invalid die-size used (must be between 1 and {max_diesize} sides)")

        # At this point we know we have valid input - roll and add dice together
        return sum(randint(1, diesize) for _ in range(number))

    def roll_with_advantage_or_disadvantage(self, advantage=False, disadvantage=False):
        """
        Base roll of d20, or 2d20, based on dis/advantage given.

        Args:
            bonus (int): The ability bonus to apply, like strength or charisma.
            advantage (bool): Roll 2d20 and use the bigger number.
            disadvantage (bool): Roll 2d20 and use the smaller number.

        Notes:
            Disadvantage and advantage cancel each other out.

        """
        if not (advantage or disadvantage) or (advantage and disadvantage):
            # normal roll, or advantage cancels disadvantage
            return self.roll("1d20")
        elif advantage:
            return max(self.roll("1d20"), self.roll("1d20"))
        else:
            return min(self.roll("1d20"), self.roll("1d20"))

    def saving_throw(
        self,
        character,
        bonus_type=Ability.STR,
        target=15,
        advantage=False,
        disadvantage=False,
        modifier=0,
    ):
        """
        A saving throw without a clear enemy to beat. In _Knave_ all unopposed saving
        throws always tries to beat 15, so (d20 + bonus + modifier) > 15.

        Args:
            character (Object): The one attempting to save themselves.
            bonus_type (enum.Ability): The ability bonus to apply, like strength or
                charisma.
            target (int, optional): Used for opposed throws (in Knave any regular
                saving through must always beat 15).
            advantage (bool, optional): Roll 2d20 and use the bigger number.
            disadvantage (bool, optional): Roll 2d20 and use the smaller number.
            modifier (int, optional): An additional +/- modifier to the roll.

        Returns:
            tuple: A tuple `(bool, str, str)`. The bool indicates if the save was passed or not.
                The second element is the quality of the roll - None (normal),
                "critical fail" and "critical success". Last element is a text detailing
                the roll, for display purposes.
        Notes:
            Advantage and disadvantage cancel each other out.

        Example:
            Trying to overcome the effects of poison, roll d20 + Constitution-bonus above 15.

        """
        bonus = getattr(character, bonus_type.value, 1)
        dice_roll = self.roll_with_advantage_or_disadvantage(advantage, disadvantage)
        if dice_roll == 1:
            quality = Ability.CRITICAL_FAILURE
        elif dice_roll == 20:
            quality = Ability.CRITICAL_SUCCESS
        else:
            quality = None
        result = dice_roll + bonus + modifier > target

        # determine text output
        rolltxt = "d20 "
        if advantage and disadvantage:
            rolltxt = "d20 (advantage canceled by disadvantage)"
        elif advantage:
            rolltxt = "|g2d20|n (advantage: picking highest) "
        elif disadvantage:
            rolltxt = "|r2d20|n (disadvantage: picking lowest) "
        bontxt = f"(+{bonus})"
        modtxt = ""
        if modifier:
            modtxt = f"+ {modifier}" if modifier > 0 else f" - {abs(modifier)}"
        qualtxt = f" ({quality.value}!)" if quality else ""

        txt = (
            f" rolled {dice_roll} on {rolltxt} "
            f"+ {bonus_type.value}{bontxt}{modtxt} vs "
            f"{target} -> |w{'|GSuccess|w' if result else '|RFail|w'}{qualtxt}|n"
        )

        return (dice_roll + bonus + modifier) > target, quality, txt

    def opposed_saving_throw(
        self,
        attacker,
        defender,
        attack_type=Ability.STR,
        defense_type=Ability.ARMOR,
        advantage=False,
        disadvantage=False,
        modifier=0,
    ):
        """
        An saving throw that tries to beat an active opposing side.

        Args:
            attacker (Character): The attacking party.
            defender (Character): The one defending.
            attack_type (str): Which ability to use in the attack, like 'strength' or 'willpower'.
                Minimum is always 1.
            defense_type (str): Which ability to defend with, in addition to 'armor'.
                Minimum is always 11 (bonus + 10 is always the defense in _Knave_).
            advantage (bool): Roll 2d20 and use the bigger number.
            disadvantage (bool): Roll 2d20 and use the smaller number.
            modifier (int): An additional +/- modifier to the roll.

        Returns:
            tuple: (bool, str, str): If the attack succeed or not. The second element is the
                quality of the roll - None (normal), "critical fail" and "critical success". Last
                element is a text that summarizes the details of the roll.
        Notes:
            Advantage and disadvantage cancel each other out.

        """
        # what is stored on the character/npc is the bonus; we add 10 to get the defense target
        defender_defense = getattr(defender, defense_type.value, 1) + 10

        result, quality, txt = self.saving_throw(
            attacker,
            bonus_type=attack_type,
            target=defender_defense,
            advantage=advantage,
            disadvantage=disadvantage,
            modifier=modifier,
        )
        txt = f"Roll vs {defense_type.value}({defender_defense}):\n{txt}"

        return result, quality, txt

    def roll_random_table(self, dieroll, table_choices):
        """
        Make a roll on a random table.

        Args:
            dieroll (str): The dice to roll, like 1d6, 1d20, 3d6 etc).
            table_choices (iterable): If a list of single elements, the die roll
                should fully encompass the table, like a 1d20 roll for a table
                with 20 elements. If each element is a tuple, the first element
                of the tuple is assumed to be a string 'X-Y' indicating the
                range of values that should match the roll.

        Returns:
            Any: The result of the random roll.

        Example:
            `roll table_choices = [('1-5', "Blue"), ('6-9': "Red"), ('10', "Purple")]`

        Notes:
            If the roll is outside of the listing, the closest edge value is used.

        """
        roll_result = self.roll(dieroll)
        if not table_choices:
            return None

        if isinstance(table_choices[0], (tuple, list)):
            # tuple with range conditional, like ('1-5', "Blue") or ('10', "Purple")
            max_range = -1
            min_range = 10**6
            for valrange, choice in table_choices:
                minval, *maxval = valrange.split("-", 1)
                minval = abs(int(minval))
                maxval = abs(int(maxval[0]) if maxval else minval)

                # we store the largest/smallest values so far in case we need to use them
                max_range = max(max_range, maxval)
                min_range = min(min_range, minval)

                if minval <= roll_result <= maxval:
                    return choice

            # if we have no result, we are outside of the range, we pick the edge values. It is also
            # possible the range contains 'gaps', but that'd be an error in the random table itself.
            if roll_result > max_range:
                return table_choices[-1][1]
            else:
                return table_choices[0][1]
        else:
            # regular list - one line per value.
            roll_result = max(1, min(len(table_choices), roll_result))
            return table_choices[roll_result - 1]

    # specific rolls / actions

    def morale_check(self, defender):
        """
        A morale check is done for NPCs/monsters. It's done with a 2d6 against
        their morale.

        Args:
            defender (NPC): The entity trying to defend its morale.

        Returns:
            bool: False if morale roll failed, True otherwise.

        """
        return self.roll("2d6") <= defender.morale

    def heal_from_rest(self, character):
        """
        A meal and a full night's rest allow for regaining 1d8 + Const bonus HP.

        Args:
            character (Character): The one resting.

        """
        character.heal(self.roll("1d8") + character.constitution)

    death_map = {
        "weakened": "strength",
        "unsteady": "dexterity",
        "sickly": "constitution",
        "addled": "intelligence",
        "rattled": "wisdom",
        "disfigured": "charisma",
    }

    def roll_death(self, character):
        """
        Happens when hitting <= 0 hp. unless dead,

        """

        result = self.roll_random_table("1d8", death_table)
        if result == "dead":
            character.at_death()
        else:
            # survives with degraded abilities (1d4 roll)
            abi = self.death_map[result]

            current_abi = getattr(character, abi)
            loss = self.roll("1d4")

            current_abi -= loss

            if current_abi < -10:
                # can't lose more - die
                character.at_death()
            else:
                # refresh health, but get permanent ability loss
                new_hp = self.roll("1d4")
                character.heal(new_hp)
                setattr(character, abi, current_abi)

                character.msg(
                    "~" * 78 + "\n|yYou survive your brush with death, "
                    f"but are |r{result.upper()}|y and permanently |rlose {loss} {abi}|y.|n\n"
                    f"|GYou recover |g{new_hp}|G health|.\n" + "~" * 78
                )


# singletons

# access rolls e.g. with rules.dice.opposed_saving_throw(...)
dice = EvAdventureRollEngine()



"""
Unittests for help code (The default help-command is tested as part of default
command test-suite).

"""

from unittest import mock

from evennia.help import filehelp
from evennia.help import utils as help_utils
from evennia.utils.test_resources import TestCase
from evennia.utils.utils import dedent


class TestParseSubtopics(TestCase):
    """
    Test the subtopic parser.

    """

    def test_parse_entry(self):
        """
        Test for subcategories

        """
        self.maxDiff = None

        entry = dedent(
            """
            Main topic text
            # subtopics
            ## foo
            Foo sub-category
            ### moo
            Foo/Moo subsub-category
            #### dum
            Foo/Moo/Dum subsubsub-category
            ## bar
            Bar subcategory
            ### moo
            Bar/Moo subcategory
        """,
            indent=0,
        )
        expected = {
            None: "Main topic text",
            "foo": {
                None: "\nFoo sub-category\n",
                "moo": {
                    None: "\nFoo/Moo subsub-category\n",
                    "dum": {
                        None: "\nFoo/Moo/Dum subsubsub-category\n",
                    },
                },
            },
            "bar": {None: "\nBar subcategory\n", "moo": {None: "\nBar/Moo subcategory"}},
        }

        actual_result = help_utils.parse_entry_for_subcategories(entry)
        self.assertEqual(expected, actual_result)

    def test_parse_single_entry(self):
        """
        Test parsing single subcategory

        """

        entry = dedent(
            """
        Main topic text
        # SUBTOPICS
        ## creating extra stuff
        Help on creating extra stuff.
        """,
            indent=0,
        )
        expected = {
            None: "Main topic text",
            "creating extra stuff": {None: "\nHelp on creating extra stuff."},
        }

        actual_result = help_utils.parse_entry_for_subcategories(entry)
        self.assertEqual(expected, actual_result)


# test filehelp system

HELP_ENTRY_DICTS = [
    {
        "key": "evennia",
        "aliases": ["ev"],
        "category": "General",
        "text": """
            Evennia is a MUD game server in Python.

            # subtopics

            ## Installation

            You'll find installation instructions on https:evennia.com

            ## Community

            There are many ways to get help and communicate with other devs!

            ### Discord

            There is also a discord channel you can find from the sidebard on evennia.com.

        """,
    },
    {
        "key": "building",
        "category": "building",
        "text": """
            Evennia comes with a bunch of default building commands. You can
            find a building tutorial in the evennia documentation.

        """,
    },
]


class TestFileHelp(TestCase):
    """
    Test the File-help system

    """

    @mock.patch("evennia.help.filehelp.variable_from_module")
    def test_file_help(self, mock_variable_from_module):
        mock_variable_from_module.return_value = HELP_ENTRY_DICTS

        # we just need anything here since we mock the load anyway
        storage = filehelp.FileHelpStorageHandler(help_file_modules=["dummypath"])
        result = storage.all()

        for inum, helpentry in enumerate(result):
            self.assertEqual(HELP_ENTRY_DICTS[inum]["key"], helpentry.key)
            self.assertEqual(HELP_ENTRY_DICTS[inum].get("aliases", []), helpentry.aliases)
            self.assertEqual(HELP_ENTRY_DICTS[inum]["category"], helpentry.help_category)
            self.assertEqual(HELP_ENTRY_DICTS[inum]["text"], helpentry.entrytext)



from anything import Something
from django.test import TestCase
from mock import Mock, mock_open, patch

from .dummyrunner_settings import (
    c_creates_button,
    c_creates_obj,
    c_digs,
    c_examines,
    c_help,
    c_idles,
    c_login,
    c_login_nodig,
    c_logout,
    c_looks,
    c_moves,
    c_moves_n,
    c_moves_s,
    c_socialize,
)

try:
    import memplot
except ImportError:
    memplot = Mock()


class TestDummyrunnerSettings(TestCase):
    def setUp(self):
        self.client = Mock()
        self.client.cid = 1
        self.client.counter = Mock(return_value=1)
        self.client.gid = "20171025161153-1"
        self.client.name = "Dummy_%s" % self.client.gid
        self.client.password = (Something,)
        self.client.start_room = "testing_room_start_%s" % self.client.gid
        self.client.objs = []
        self.client.exits = []

    def clear_client_lists(self):
        self.client.objs = []
        self.client.exits = []

    def test_c_login(self):
        self.assertEqual(
            c_login(self.client),
            (
                Something,  # create
                "yes",  # confirm creation
                Something,  # connect
                "dig %s" % self.client.start_room,
                "teleport %s" % self.client.start_room,
                "py from evennia.server.profiling.dummyrunner import DummyRunnerCmdSet;"
                "self.cmdset.add(DummyRunnerCmdSet, persistent=False)",
            ),
        )

    def test_c_login_no_dig(self):
        cmd1, cmd2 = c_login_nodig(self.client)
        self.assertTrue(cmd1.startswith("create " + self.client.name + " "))
        self.assertTrue(cmd2.startswith("connect " + self.client.name + " "))

    def test_c_logout(self):
        self.assertEqual(c_logout(self.client), ("quit",))

    def perception_method_tests(self, func, verb, alone_suffix=""):
        self.assertEqual(func(self.client), ("%s%s" % (verb, alone_suffix),))
        self.client.exits = ["exit1", "exit2"]
        self.assertEqual(func(self.client), ["%s exit1" % verb, "%s exit2" % verb])
        self.client.objs = ["foo", "bar"]
        self.assertEqual(func(self.client), ["%s foo" % verb, "%s bar" % verb])
        self.clear_client_lists()

    def test_c_looks(self):
        self.perception_method_tests(c_looks, "look")

    def test_c_examines(self):
        self.perception_method_tests(c_examines, "examine", " me")

    def test_idles(self):
        self.assertEqual(c_idles(self.client), ("idle", "idle"))

    def test_c_help(self):
        self.assertEqual(
            c_help(self.client),
            ("help", "dummyrunner_echo_response"),
        )

    def test_c_digs(self):
        self.assertEqual(c_digs(self.client), ("dig/tel testing_room_1 = exit_1, exit_1",))
        self.assertEqual(self.client.exits, ["exit_1", "exit_1"])
        self.clear_client_lists()

    def test_c_creates_obj(self):
        objname = "testing_obj_1"
        self.assertEqual(
            c_creates_obj(self.client),
            (
                "create %s" % objname,
                'desc %s = "this is a test object' % objname,
                "set %s/testattr = this is a test attribute value." % objname,
                "set %s/testattr2 = this is a second test attribute." % objname,
            ),
        )
        self.assertEqual(self.client.objs, [objname])
        self.clear_client_lists()

    def test_c_creates_button(self):
        objname = "testing_button_1"
        typeclass_name = "contrib.tutorial_examples.red_button.RedButton"
        self.assertEqual(
            c_creates_button(self.client),
            ("create %s:%s" % (objname, typeclass_name), "desc %s = test red button!" % objname),
        )
        self.assertEqual(self.client.objs, [objname])
        self.clear_client_lists()

    def test_c_socialize(self):
        self.assertEqual(
            c_socialize(self.client),
            (
                "pub Hello!",
                "say Yo!",
                "emote stands looking around.",
            ),
        )

    def test_c_moves(self):
        self.assertEqual(c_moves(self.client), ("look",))
        self.client.exits = ["south", "north"]
        self.assertEqual(c_moves(self.client), ["south", "north"])
        self.clear_client_lists()

    def test_c_move_n(self):
        self.assertEqual(c_moves_n(self.client), ("north",))

    def test_c_move_s(self):
        self.assertEqual(c_moves_s(self.client), ("south",))


class TestMemPlot(TestCase):
    @patch.object(memplot, "_idmapper")
    @patch.object(memplot, "os")
    @patch.object(memplot, "open", new_callable=mock_open, create=True)
    @patch.object(memplot, "time")
    @patch("evennia.utils.idmapper.models.SharedMemoryModel.flush_from_cache", new=Mock())
    def test_memplot(self, mock_time, mocked_open, mocked_os, mocked_idmapper):
        if isinstance(memplot, Mock):
            return
        from evennia.utils.create import create_script

        mocked_idmapper.cache_size.return_value = (9, 5000)
        mock_time.time = Mock(return_value=6000.0)
        script = create_script(memplot.Memplot)
        script.db.starttime = 0.0
        mocked_os.popen.read.return_value = 5000.0
        script.at_repeat()
        handle = mocked_open()
        handle.write.assert_called_with("100.0, 0.001, 0.001, 9\n")
        script.stop()



"""
Evscaperoom - Griatch 2019

"""

from . import commands  # noqa
from . import menu  # noqa
from . import objects  # noqa
from . import room  # noqa
from . import scripts  # noqa
from . import state  # noqa
from . import tests  # noqa
from . import utils  # noqa



"""
Test characters.

"""

from evennia.utils import create
from evennia.utils.test_resources import BaseEvenniaTest

from ..characters import EvAdventureCharacter


class TestCharacters(BaseEvenniaTest):
    def setUp(self):
        super().setUp()
        self.character = create.create_object(EvAdventureCharacter, key="testchar")

    def test_abilities(self):
        self.character.strength += 2
        self.assertEqual(self.character.strength, 3)

    def test_heal(self):
        """Make sure we don't heal too much"""
        self.character.hp = 0
        self.character.hp_max = 8

        self.character.heal(1)
        self.assertEqual(self.character.hp, 1)
        self.character.heal(100)
        self.assertEqual(self.character.hp, 8)

    def test_at_damage(self):
        self.character.hp = 8
        self.character.at_damage(5)
        self.assertEqual(self.character.hp, 3)

    def test_at_pay(self):
        self.character.coins = 100

        result = self.character.at_pay(60)
        self.assertEqual(result, 60)
        self.assertEqual(self.character.coins, 40)

        # can't get more coins than we have
        result = self.character.at_pay(100)
        self.assertEqual(result, 40)
        self.assertEqual(self.character.coins, 0)



"""
Search and multimatch handling

This module allows for overloading two functions used by Evennia's
search functionality:

    at_search_result:
        This is called whenever a result is returned from an object
        search (a common operation in commands).  It should (together
        with at_multimatch_input below) define some way to present and
        differentiate between multiple matches (by default these are
        presented as 1-ball, 2-ball etc)
    at_multimatch_input:
        This is called with a search term and should be able to
        identify if the user wants to separate a multimatch-result
        (such as that from a previous search). By default, this
        function understands input on the form 1-ball, 2-ball etc as
        indicating that the 1st or 2nd match for "ball" should be
        used.

This module is not called by default, to use it, add the following
line to your settings file:

    SEARCH_AT_RESULT = "server.conf.at_search.at_search_result"

"""


def at_search_result(matches, caller, query="", quiet=False, **kwargs):
    """
    This is a generic hook for handling all processing of a search
    result, including error reporting.

    Args:
        matches (list): This is a list of 0, 1 or more typeclass instances,
            the matched result of the search. If 0, a nomatch error should
            be echoed, and if >1, multimatch errors should be given. Only
            if a single match should the result pass through.
        caller (Object): The object performing the search and/or which should
        receive error messages.
    query (str, optional): The search query used to produce `matches`.
        quiet (bool, optional): If `True`, no messages will be echoed to caller
            on errors.

    Keyword Args:
        nofound_string (str): Replacement string to echo on a notfound error.
        multimatch_string (str): Replacement string to echo on a multimatch error.

    Returns:
        processed_result (Object or None): This is always a single result
            or `None`. If `None`, any error reporting/handling should
            already have happened.

    """



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-06 17:31


import django.contrib.auth.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("accounts", "0005_auto_20160905_0902")]

    operations = [
        migrations.AlterField(
            model_name="accountdb",
            name="db_attributes",
            field=models.ManyToManyField(
                help_text="attributes on this object. An attribute can hold any pickle-able python object (see docs for special cases).",
                to="typeclasses.Attribute",
            ),
        ),
        migrations.AlterField(
            model_name="accountdb",
            name="db_tags",
            field=models.ManyToManyField(
                help_text="tags on this object. Tags are simple string markers to identify, group and alias objects.",
                to="typeclasses.Tag",
            ),
        ),
        migrations.AlterField(
            model_name="accountdb",
            name="username",
            field=models.CharField(
                error_messages={"unique": "A user with that username already exists."},
                help_text="Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.",
                max_length=150,
                unique=True,
                validators=[django.contrib.auth.validators.ASCIIUsernameValidator()],
                verbose_name="username",
            ),
        ),
    ]



"""
Testing custom game time

"""

# Testing custom_gametime
from mock import Mock, patch

from evennia.utils.test_resources import BaseEvenniaTest

from .. import custom_gametime


def _testcallback():
    pass


@patch("evennia.utils.gametime.gametime", new=Mock(return_value=2975000898.46))
class TestCustomGameTime(BaseEvenniaTest):
    def tearDown(self):
        if hasattr(self, "timescript"):
            self.timescript.stop()

    def test_time_to_tuple(self):
        self.assertEqual(custom_gametime.time_to_tuple(10000, 34, 2, 4, 6, 1), (294, 2, 0, 0, 0, 0))
        self.assertEqual(custom_gametime.time_to_tuple(10000, 3, 3, 4), (3333, 0, 0, 1))
        self.assertEqual(custom_gametime.time_to_tuple(100000, 239, 24, 3), (418, 4, 0, 2))

    def test_gametime_to_realtime(self):
        self.assertEqual(custom_gametime.gametime_to_realtime(days=2, mins=4), 86520.0)
        self.assertEqual(
            custom_gametime.gametime_to_realtime(format=True, days=2), (0, 0, 0, 1, 0, 0, 0)
        )

    def test_realtime_to_gametime(self):
        self.assertEqual(custom_gametime.realtime_to_gametime(days=3, mins=34), 349680.0)
        self.assertEqual(
            custom_gametime.realtime_to_gametime(days=3, mins=34, format=True),
            (0, 0, 0, 4, 1, 8, 0),
        )
        self.assertEqual(
            custom_gametime.realtime_to_gametime(format=True, days=3, mins=4), (0, 0, 0, 4, 0, 8, 0)
        )

    def test_custom_gametime(self):
        self.assertEqual(custom_gametime.custom_gametime(), (102, 5, 2, 6, 21, 8, 18))
        self.assertEqual(custom_gametime.custom_gametime(absolute=True), (102, 5, 2, 6, 21, 8, 18))

    def test_real_seconds_until(self):
        self.assertEqual(
            custom_gametime.real_seconds_until(year=2300, month=12, day=7), 31911667199.77
        )

    def test_schedule(self):
        self.timescript = custom_gametime.schedule(_testcallback, repeat=True, min=5, sec=0)
        self.assertEqual(self.timescript.interval, 1700.7699999809265)



"""
Dummy client runner

This module implements a stand-alone launcher for stress-testing
an Evennia game. It will launch any number of fake clients. These
clients will log into the server and start doing random operations.
Customizing and weighing these operations differently depends on
which type of game is tested. The module contains a testing module
for plain Evennia.

Please note that you shouldn't run this on a production server!
Launch the program without any arguments or options to see a
full step-by-step setup help.

Basically (for testing default Evennia):

 - Use an empty/testing database.
 - set PERMISSION_ACCOUNT_DEFAULT = "Builder"
 - start server, eventually with profiling active
 - launch this client runner

If you want to customize the runner's client actions
(because you changed the cmdset or needs to better
match your use cases or add more actions), you can
change which actions by adding a path to

   DUMMYRUNNER_ACTIONS_MODULE = <path.to.your.module>

in your settings. See utils.dummyrunner_actions.py
for instructions on how to define this module.

"""


import random
import sys
import time
from argparse import ArgumentParser

import django
from twisted.conch import telnet
from twisted.internet import protocol, reactor
from twisted.internet.task import LoopingCall

django.setup()
import evennia  # noqa

evennia._init()

from django.conf import settings  # noqa

from evennia.commands.cmdset import CmdSet  # noqa
from evennia.commands.command import Command  # noqa
from evennia.utils import mod_import, time_format  # noqa
from evennia.utils.ansi import strip_ansi  # noqa

# Load the dummyrunner settings module

DUMMYRUNNER_SETTINGS = mod_import(settings.DUMMYRUNNER_SETTINGS_MODULE)
if not DUMMYRUNNER_SETTINGS:
    raise IOError(
        "Error: Dummyrunner could not find settings file at %s"
        % settings.DUMMYRUNNER_SETTINGS_MODULE
    )
IDMAPPER_CACHE_MAXSIZE = settings.IDMAPPER_CACHE_MAXSIZE

DATESTRING = "%Y%m%d%H%M%S"
CLIENTS = []

# Settings

# number of clients to launch if no input is given on command line
NCLIENTS = 1
# time between each 'tick', in seconds, if not set on command
# line. All launched clients will be called upon to possibly do an
# action with this frequency.
TIMESTEP = DUMMYRUNNER_SETTINGS.TIMESTEP
# chance of a client performing an action, per timestep. This helps to
# spread out usage randomly, like it would be in reality.
CHANCE_OF_ACTION = DUMMYRUNNER_SETTINGS.CHANCE_OF_ACTION
# spread out the login action separately, having many accounts create accounts
# and connect simultaneously is generally unlikely.
CHANCE_OF_LOGIN = DUMMYRUNNER_SETTINGS.CHANCE_OF_LOGIN
# Port to use, if not specified on command line
TELNET_PORT = DUMMYRUNNER_SETTINGS.TELNET_PORT or settings.TELNET_PORTS[0]
#
NCONNECTED = 0  # client has received a connection
NLOGIN_SCREEN = 0  # client has seen the login screen (server responded)
NLOGGING_IN = 0  # client starting login procedure
NLOGGED_IN = 0  # client has authenticated and logged in

# time when all clients have logged_in
TIME_ALL_LOGIN = 0
# actions since all logged in
TOTAL_ACTIONS = 0
TOTAL_LAG_MEASURES = 0
# lag per 30s for all logged in
TOTAL_LAG = 0
TOTAL_LAG_IN = 0
TOTAL_LAG_OUT = 0


INFO_STARTING = """
    Dummyrunner starting using {nclients} dummy account(s). If you don't see
    any connection messages, make sure that the Evennia server is
    running.

    TELNET_PORT = {port}
    IDMAPPER_CACHE_MAXSIZE = {idmapper_cache_size} MB
    TIMESTEP = {timestep} (rate {rate}/s)
    CHANCE_OF_LOGIN = {chance_of_login}% per time step
    CHANCE_OF_ACTION = {chance_of_action}% per time step
    -> avg rate (per client, after login): {avg_rate} cmds/s
    -> total avg rate (after login): {avg_rate_total} cmds/s

    Use Ctrl-C (or Cmd-C) to stop/disconnect all clients.

    """

ERROR_NO_MIXIN = """
    Error: Evennia is not set up for dummyrunner. Before starting the
    server, make sure to include the following at *the end* of your
    settings file (remove when not using dummyrunner!):

        from evennia.server.profiling.settings_mixin import *

    This will change the settings in the following way:
        - change PERMISSION_ACCOUNT_DEFAULT to 'Developer' to allow clients
          to test all commands
        - change PASSWORD_HASHERS to use a faster (but less safe) algorithm
          when creating large numbers of accounts at the same time
        - set LOGIN_THROTTLE/CREATION_THROTTLE=None to disable it

    If you don't want to use the custom settings of the mixin for some
    reason, you can change their values manually after the import, or
    add DUMMYRUNNER_MIXIN=True to your settings file to avoid this
    error completely.

    Warning: Don't run dummyrunner on a production database! It will
    create a lot of spammy objects and accounts!
    """


ERROR_FEW_ACTIONS = """
    Dummyrunner settings error: The ACTIONS tuple is too short: it must
    contain at least login- and logout functions.
    """


HELPTEXT = """
DO NOT RUN THIS ON A PRODUCTION SERVER! USE A CLEAN/TESTING DATABASE!

This stand-alone program launches dummy telnet clients against a
running Evennia server. The idea is to mimic real accounts logging in
and repeatedly doing resource-heavy commands so as to stress test the
game. It uses the default command set to log in and issue commands, so
if that was customized, some of the functionality will not be tested
(it will not fail, the commands will just not be recognized).  The
running clients will create new objects and rooms all over the place
as part of their running, so using a clean/testing database is
strongly recommended.

Setup:
  1) setup a fresh/clean database (if using sqlite, just safe-copy
     away your real evennia.db3 file and create a new one with
     `evennia migrate`)
  2) in server/conf/settings.py, add

        PERMISSION_ACCOUNT_DEFAULT="Builder"

     This is so that the dummy accounts can test building operations.
     You can also customize the dummyrunner by modifying a setting
     file specified by DUMMYRUNNER_SETTINGS_MODULE

  3) Start Evennia like normal, optionally with profiling (--profile)
  4) Run this dummy runner via the evennia launcher:

        evennia --dummyrunner <nr_of_clients>

  5) Log on and determine if game remains responsive despite the
     heavier load. Note that if you activated profiling, there is a
     considerate additional overhead from the profiler too so you
     should usually not consider game responsivity when using the
     profiler at the same time.
  6) If you use profiling, let the game run long enough to gather
     data, then stop the server cleanly using evennia stop or @shutdown.
     @shutdown. The profile appears as
     server/logs/server.prof/portal.prof (see Python's manual on
     cProfiler).

Notes:

The dummyrunner tends to create a lot of accounts all at once, which is
a very heavy operation. This is not a realistic use-case - what you want
to test is performance during run. A large
number of clients here may lock up the client until all have been
created. It may be better to connect multiple dummyrunners instead of
starting one single one with a lot of accounts. Exactly what this number
is depends on your computer power. So start with 10-20 clients and increase
until you see the initial login slows things too much.

"""


class CmdDummyRunnerEchoResponse(Command):
    """
    Dummyrunner command measuring the round-about response time
    from sending to receiving a result.

    Usage:
        dummyrunner_echo_response <timestamp>

    Responds with
        dummyrunner_echo_response:<timestamp>,<current_time>

    The dummyrunner will send this and then compare the send time
    with the receive time on both ends.

    """

    key = "dummyrunner_echo_response"

    def func(self):
        # returns (dummy_client_timestamp,current_time)
        self.msg(f"dummyrunner_echo_response:{self.args},{time.time()}")
        if self.caller.account.is_superuser:
            print(f"cmddummyrunner lag in: {time.time() - float(self.args)}s")


class DummyRunnerCmdSet(CmdSet):
    """
    Dummyrunner injected cmdset.

    """

    def at_cmdset_creation(self):
        self.add(CmdDummyRunnerEchoResponse())


# ------------------------------------------------------------
# Helper functions
# ------------------------------------------------------------


ICOUNT = 0


def idcounter():
    """
    Makes unique ids.

    Returns:
        str: A globally unique id.

    """
    global ICOUNT
    ICOUNT += 1
    return str("{:03d}".format(ICOUNT))


GCOUNT = 0


def gidcounter():
    """
    Makes globally unique ids.

    Returns:
        count (int); A globally unique counter.

    """
    global GCOUNT
    GCOUNT += 1
    return "%s_%s" % (time.strftime(DATESTRING), GCOUNT)


def makeiter(obj):
    """
    Makes everything iterable.

    Args:
        obj (any): Object to turn iterable.

    Returns:
        iterable (iterable): An iterable object.
    """
    return obj if hasattr(obj, "__iter__") else [obj]


# ------------------------------------------------------------
# Client classes
# ------------------------------------------------------------


class DummyClient(telnet.StatefulTelnetProtocol):
    """
    Handles connection to a running Evennia server,
    mimicking a real account by sending commands on
    a timer.

    """

    def report(self, text, clientkey):
        pad = " " * (25 - len(text))
        tim = round(time.time() - self.connection_timestamp)
        print(
            f"{text} {clientkey}{pad}\t"
            f"conn: {NCONNECTED} -> "
            f"welcome screen: {NLOGIN_SCREEN} -> "
            f"authing: {NLOGGING_IN} -> "
            f"loggedin/tot: {NLOGGED_IN}/{NCLIENTS} (after {tim}s)"
        )

    def connectionMade(self):
        """
        Called when connection is first established.

        """
        global NCONNECTED
        # public properties
        self.cid = idcounter()
        self.key = f"Dummy-{self.cid}"
        self.gid = f"{time.strftime(DATESTRING)}_{self.cid}"
        self.istep = 0
        self.exits = []  # exit names created
        self.objs = []  # obj names created
        self.connection_timestamp = time.time()
        self.connection_attempt = 0
        self.action_started = 0

        self._connected = False
        self._loggedin = False
        self._logging_out = False
        self._ready = False
        self._report = ""
        self._cmdlist = []  # already stepping in a cmd definition
        self._login = self.factory.actions[0]
        self._logout = self.factory.actions[1]
        self._actions = self.factory.actions[2:]

        reactor.addSystemEventTrigger("before", "shutdown", self.logout)

        NCONNECTED += 1
        self.report("-> connected", self.key)

        reactor.callLater(30, self._retry_welcome_screen)

    def _retry_welcome_screen(self):
        if not self._connected and not self._ready:
            # we have connected but not received anything for 30s.
            # (unclear why this would be - overload?)
            # try sending a look to get something to start with
            self.report("?? retrying welcome screen", self.key)
            self.sendLine(bytes("look", "utf-8"))
            # make sure to check again later
            reactor.callLater(30, self._retry_welcome_screen)

    def _print_statistics(self):
        global TIME_ALL_LOGIN, TOTAL_ACTIONS
        global TOTAL_LAG, TOTAL_LAG_MEASURES, TOTAL_LAG_IN, TOTAL_LAG_OUT

        tim = time.time() - TIME_ALL_LOGIN
        avgrate = round(TOTAL_ACTIONS / tim)
        lag = TOTAL_LAG / (TOTAL_LAG_MEASURES or 1)
        lag_in = TOTAL_LAG_IN / (TOTAL_LAG_MEASURES or 1)
        lag_out = TOTAL_LAG_OUT / (TOTAL_LAG_MEASURES or 1)

        TOTAL_ACTIONS = 0
        TOTAL_LAG = 0
        TOTAL_LAG_IN = 0
        TOTAL_LAG_OUT = 0
        TOTAL_LAG_MEASURES = 0
        TIME_ALL_LOGIN = time.time()

        print(
            f".. running 30s average: ~{avgrate} actions/s "
            f"lag: {lag:.2}s (in: {lag_in:.2}s, out: {lag_out:.2}s)"
        )

        reactor.callLater(30, self._print_statistics)

    def dataReceived(self, data):
        """
        Called when data comes in over the protocol. We wait to start
        stepping until the server actually responds

        Args:
            data (str): Incoming data.

        """
        global NLOGIN_SCREEN, NLOGGED_IN, NLOGGING_IN, NCONNECTED
        global TOTAL_ACTIONS, TIME_ALL_LOGIN
        global TOTAL_LAG, TOTAL_LAG_MEASURES, TOTAL_LAG_IN, TOTAL_LAG_OUT

        if not data.startswith(b"\xff"):
            # regular text, not a telnet command

            if NCLIENTS == 1:
                print("dummy-client sees:", str(data, "utf-8"))

            if not self._connected:
                # waiting for connection
                # wait until we actually get text back (not just telnet
                # negotiation)
                # start client tick
                d = LoopingCall(self.step)
                df = max(abs(TIMESTEP * 0.001), min(TIMESTEP / 10, 0.5))
                # dither next attempt with random time
                timestep = TIMESTEP + (-df + (random.random() * df))
                d.start(timestep, now=True).addErrback(self.error)
                self.connection_attempt += 1

                self._connected = True
                NLOGIN_SCREEN += 1
                NCONNECTED -= 1
                self.report("<- server sent login screen", self.key)

            elif self._loggedin:
                if not self._ready:
                    # logged in, ready to run
                    NLOGGED_IN += 1
                    NLOGGING_IN -= 1
                    self._ready = True
                    self.report("== logged in", self.key)
                    if NLOGGED_IN == NCLIENTS and not TIME_ALL_LOGIN:
                        # all are logged in! We can start collecting statistics
                        print(".. All clients connected and logged in!")
                        TIME_ALL_LOGIN = time.time()
                        reactor.callLater(30, self._print_statistics)

                elif TIME_ALL_LOGIN:
                    TOTAL_ACTIONS += 1

                    try:
                        data = strip_ansi(str(data, "utf-8").strip())
                        if data.startswith("dummyrunner_echo_response:"):
                            # handle special lag-measuring command. This returns
                            # dummyrunner_echo_response:<starttime>,<midpointtime>
                            now = time.time()
                            _, data = data.split(":", 1)
                            start_time, mid_time = (float(part) for part in data.split(",", 1))
                            lag_in = mid_time - start_time
                            lag_out = now - mid_time
                            total_lag = now - start_time  # full round-about time

                            TOTAL_LAG += total_lag
                            TOTAL_LAG_IN += lag_in
                            TOTAL_LAG_OUT += lag_out
                            TOTAL_LAG_MEASURES += 1
                    except Exception:
                        pass

    def connectionLost(self, reason):
        """
        Called when loosing the connection.

        Args:
            reason (str): Reason for loosing connection.

        """
        if not self._logging_out:
            self.report("XX lost connection", self.key)

    def error(self, err):
        """
        Error callback.

        Args:
            err (Failure): Error instance.
        """
        print(err)

    def counter(self):
        """
        Produces a unique id, also between clients.

        Returns:
            counter (int): A unique counter.

        """
        return gidcounter()

    def logout(self):
        """
        Causes the client to log out of the server. Triggered by ctrl-c signal.

        """
        self._logging_out = True
        cmd = self._logout(self)[0]
        self.report(f"-> logout/disconnect ({self.istep} actions)", self.key)
        self.sendLine(bytes(cmd, "utf-8"))

    def step(self):
        """
        Perform a step. This is called repeatedly by the runner and
        causes the client to issue commands to the server.  This holds
        all "intelligence" of the dummy client.

        """
        global NLOGGING_IN, NLOGIN_SCREEN

        rand = random.random()

        if not self._cmdlist:
            # no commands ready. Load some.

            if not self._loggedin:
                if rand < CHANCE_OF_LOGIN or NLOGGING_IN < 10:
                    # lower rate of logins, but not below 1 / s
                    # get the login commands
                    self._cmdlist = list(makeiter(self._login(self)))
                    NLOGGING_IN += 1  # this is for book-keeping
                    NLOGIN_SCREEN -= 1
                    self.report("-> create/login", self.key)
                    self._loggedin = True
                else:
                    # no login yet, so cmdlist not yet set
                    return
            else:
                # we always pick a cumulatively random function
                crand = random.random()
                cfunc = [func for (cprob, func) in self._actions if cprob >= crand][0]
                self._cmdlist = list(makeiter(cfunc(self)))

        # at this point we always have a list of commands
        if rand < CHANCE_OF_ACTION:
            # send to the game
            cmd = str(self._cmdlist.pop(0))

            if cmd.startswith("dummyrunner_echo_response"):
                # we need to set the timer element as close to
                # the send as possible
                cmd = cmd.format(timestamp=time.time())

            self.sendLine(bytes(cmd, "utf-8"))
            self.action_started = time.time()
            self.istep += 1

            if NCLIENTS == 1:
                print(f"dummy-client sent: {cmd}")


class DummyFactory(protocol.ReconnectingClientFactory):
    protocol = DummyClient
    initialDelay = 1
    maxDelay = 1
    noisy = False

    def __init__(self, actions):
        "Setup the factory base (shared by all clients)"
        self.actions = actions


# ------------------------------------------------------------
# Access method:
# Starts clients and connects them to a running server.
# ------------------------------------------------------------


def start_all_dummy_clients(nclients):
    """
    Initialize all clients, connect them and start to step them

    Args:
        nclients (int): Number of dummy clients to connect.

    """
    global NCLIENTS
    NCLIENTS = int(nclients)
    actions = DUMMYRUNNER_SETTINGS.ACTIONS

    if len(actions) < 2:
        print(ERROR_FEW_ACTIONS)
        return

    # make sure the probabilities add up to 1
    pratio = 1.0 / sum(tup[0] for tup in actions[2:])
    flogin, flogout, probs, cfuncs = (
        actions[0],
        actions[1],
        [tup[0] * pratio for tup in actions[2:]],
        [tup[1] for tup in actions[2:]],
    )
    # create cumulative probabilies for the random actions
    cprobs = [sum(v for i, v in enumerate(probs) if i <= k) for k in range(len(probs))]
    # rebuild a new, optimized action structure
    actions = (flogin, flogout) + tuple(zip(cprobs, cfuncs))

    # setting up all clients (they are automatically started)
    factory = DummyFactory(actions)
    for i in range(NCLIENTS):
        reactor.connectTCP("127.0.0.1", TELNET_PORT, factory)
    # start reactor
    reactor.run()


# ------------------------------------------------------------
# Command line interface
# ------------------------------------------------------------


if __name__ == "__main__":
    try:
        settings.DUMMYRUNNER_MIXIN
    except AttributeError:
        print(ERROR_NO_MIXIN)
        sys.exit()

    # parsing command line with default vals
    parser = ArgumentParser(description=HELPTEXT)
    parser.add_argument(
        "-N", nargs=1, default=1, dest="nclients", help="Number of clients to start"
    )

    args = parser.parse_args()
    nclients = int(args.nclients[0])

    print(
        INFO_STARTING.format(
            nclients=nclients,
            port=TELNET_PORT,
            idmapper_cache_size=IDMAPPER_CACHE_MAXSIZE,
            timestep=TIMESTEP,
            rate=1 / TIMESTEP,
            chance_of_login=CHANCE_OF_LOGIN * 100,
            chance_of_action=CHANCE_OF_ACTION * 100,
            avg_rate=(1 / TIMESTEP) * CHANCE_OF_ACTION,
            avg_rate_total=(1 / TIMESTEP) * CHANCE_OF_ACTION * nclients,
        )
    )

    # run the dummyrunner
    TIME_START = t0 = time.time()
    start_all_dummy_clients(nclients=nclients)
    ttot = time.time() - t0

    # output runtime
    print("... dummy client runner stopped after %s." % time_format(ttot, style=3))
